"""
方向电荷差值的福井函数
"""
from pywfn.base import Mol
from pywfn.atomprop import dirProps
import numpy as np

class Calculator:
    def __init__(self,moln:Mol,mol0:Mol,molp:Mol):
        self.mols=[moln,mol0,molp]
        self.vects:list[np.ndarray]=None
        self.atoms:list[int]=None
        self.cals=[dirProps.Calculator(mol) for mol in self.mols]
        
    
    def calculate(self):
        for c,cal in enumerate(self.cals):
            chgs=cal.calculate()
            self.chgs[:,c]=chgs


"""
计算分子属性
"""
from functools import cached_property
from pywfn import base
from pywfn.utils import printer

# 记录TCE的HOMO能量
TCEs={
    '6-31G*':-0.33520
}

class MolProp:
    def __init__(self,mol:"base.Mol") -> None:
        self.mol=mol
    
    @cached_property
    def hlIdx(self):
        """homo和lomo的索引"""
        oets=self.mol.obtElcts
        for i,e in enumerate(oets):
            if e==0:
                return i-1,i

    @cached_property
    def homo(self):
        """HOMO轨道的能量"""
        h,l=self.hlIdx
        return self.mol.Eigenvalues[h]
    
    @cached_property
    def lomo(self):
        """HOMO轨道的能量"""
        h,l=self.hlIdx
        return self.mol.Eigenvalues[l]
    
    @cached_property
    def ecp(self):
        """化学势"""
        return (self.homo+self.lomo)/2
    
    @cached_property
    def en(self):
        """电负性"""
        return -self.ecp

    @cached_property
    def ch(self):
        """化学硬度"""
        return self.lomo-self.homo
    
    @cached_property
    def cs(self):
        """化学软度"""
        return 1/self.ch
    
    @cached_property
    def ei(self):
        """亲点指数"""
        return self.ecp**2/self.ch
    
    @cached_property
    def N(self):
        """亲和反应指标"""
        basi=self.mol.basis.name

        return self.homo-TCEs[basi]

    def props(self):
        """以字典的形式返回分子属性"""
        props = [
            ['HOMO',self.homo],
            ['LOMO',self.lomo],
            ['化学势',self.ecp],
            ['电负性',self.en],
            ['化学硬度',self.ch],
            ['化学软度',self.cs],
            ['亲点指数',self.ei],
            ['亲和反应指标',self.N]
        ]
        for k,v in props:
            printer.info(f'{k:{chr(12288)}<6}\t{v:>10.6f}')


"""
计算分子的芳香性
"""
from pywfn.base import Mol
from pywfn.atomprop import charge
from pywfn.bondprop import bondOrder
import numpy as np

class Calculator:
    def __init__(self, mol:Mol):
        self.mol = mol
        self.caler=bondOrder.Calculator(mol)
    
    def calculate(self,ratio:float=0.5):
        result=self.caler.piOrder()
        orders=result[:,-1]
        mean=np.mean(orders)
        stds=np.std(orders)
        return ratio*mean-(1-ratio)*stds


"""
生成gjf文件
"""
import os
from pathlib import Path

class Tool:
    def __init__(self,path:str) -> None:
        self.path=Path(path)
        self.coordStr=None
        self.pwd=Path(__file__).parent
        self._chk=None
        with open(os.path.join(self.pwd.parent,'data\\gjfTemplate.txt'),'r',encoding='utf-8') as f:
            self.template=f.read()

    def set_coord(self,coord:list[list[str]]):
        """写入坐标"""
        lines=[]
        for symbol,x,y,z in coord:
            x,y,z=float(x),float(y),float(z)
            line=f' {symbol}'+f'{x:.8f}'.rjust(14,' ')+f'{y:.8f}'.rjust(14,' ')+f'{z:.8f}'.rjust(14,' ')
            lines.append(line)
        self.coordStr='\n'.join(lines)
    
    def set_chk(self,path):
        self._chk=path

    def save(self):
        content=self.template.replace('<CHK>',self._chk).replace('<COORD>',self.coordStr)
        with open(f'{self.path}','w',encoding='utf-8') as f:
            f.write(content)


"""
根据三个分子的Mulliken电荷计算分子福井函数
"""
from pywfn.base import Mol
from pywfn.atomprop import charge
import numpy as np

class Calculator:
    def __init__(self,moln:Mol,mol0:Mol,molp:Mol) -> None:
        self.mols=[moln,mol0,molp]
        self.cals=[charge.Calculator(mol) for mol in self.mols]
        self.natm=len(mol0.atoms)
        self.chgs=np.zeros(shape=(self.natm,3)) # 记录所有原子的电荷
        self.dchg=np.zeros(shape=(self.natm,2)) # 记录电荷差值
    
    def calculate(self):
        for c,cal in enumerate(self.cals):
            chgs=cal.calculate()
            self.chgs[:,c]=chgs
        self.dchg[:,0]=-(self.chgs[:,0]-self.chgs[:,1])
        self.dchg[:,1]=-(self.chgs[:,1]-self.chgs[:,2])
        return self.dchg


"""
公用的工具函数
"""
import numpy as np

from pywfn.base import Atom
from pywfn.maths import vector_angle
from pywfn.utils import printer

def CM2PM(CM,orbital:list[int],oe:int)->np.ndarray:
    """
    根据系数矩阵构建密度矩阵
    CM:系数矩阵,如果是开壳层的话,列数是行数的两倍[n,n]/[n,2n]
    n:分子轨道占据电子数
    """
    PMs=CM2PMs(CM,orbital,oe)
    return np.sum(PMs,axis=0)

def CM2PMs(CM,orbital:list[int],oe:int):
    """
    构建三维密度矩阵，不要空轨道，形状为[占据轨道数,原子轨道数,原子轨道数]
    """
    A=(CM[:,orbital].T)[:,:,np.newaxis]
    B=(CM[:,orbital].T)[:,np.newaxis,:]
    return A@B*oe #用矩阵乘法的形式直接构建矩阵可比逐元素计算快多了

def judgeOrbital(centerAtom:Atom,aroundAtom:Atom,orbital:int,normal)->int:
    """
    判断一个轨道是否为π轨道,几何方法
    centerAtom,aroundAtom:原子对象
    orbital:分子轨道序数
    normal:原子法向量[其实可以是任意方向]
    """
    if aroundAtom.symbol=='H' or centerAtom.symbol=='H':
        return 0
    # 1. 根据s轨道和p轨道的贡献
    sContCenter=centerAtom.get_sCont(orbital)
    sContAround=aroundAtom.get_sCont(orbital)
    if sContCenter>0.01 or sContAround>0.01:
        return 0
    # 2. p轨道的方向要处在垂直分子平面方向
    cenDir=centerAtom.get_obtWay(orbital) #中心原子的p轨道方向
    aroDir=aroundAtom.get_obtWay(orbital)
    if cenDir is None or aroDir is None:
        return 0
    centerAngle=vector_angle(cenDir,normal)
    aroundAngle=vector_angle(aroDir,normal)
    if abs(0.5-centerAngle)<0.3 or abs(0.5-aroundAngle)<0.3:
        return 0
    # 以上两个条件都满足的可以认为是π轨道
    if (0.5-centerAngle)*(0.5-aroundAngle)>0:
        return 1
    else:
        return -1

def judgeOrbital_(centerAtom:Atom,aroundAtom:Atom,orbital:int,normal)-> int:
    """判断某一个键的分子轨道是不是π轨道,成键返回1，反键返回-1，否则返回0"""
    

    centerTs=centerAtom.pLayersTs(orbital)
    aroundTs=aroundAtom.pLayersTs(orbital)
    centerPs=[np.array(centerTs[i:i+3]) for i in range(0,len(centerTs),3)]
    aroundPs=[np.array(aroundTs[i:i+3]) for i in range(0,len(aroundTs),3)]
    normal=centerAtom.get_Normal(aroundAtom)
    centerPs_=centerAtom.get_pProj(normal, orbital) # 先计算投影
    aroundPs_=aroundAtom.get_pProj(normal, orbital)
    centerPs,centerPs_=np.array(centerPs),np.array(centerPs_) # 将系数转为数组
    aroundPs,aroundPs_=np.array(aroundPs),np.array(aroundPs_)
    centerPs,centerPs_=centerPs.sum(axis=0),centerPs_.sum(axis=0) # 将不同组的p求和
    aroundPs,aroundPs_=aroundPs.sum(axis=0),aroundPs_.sum(axis=0)
    centerL,centerL_=np.linalg.norm(centerPs),np.linalg.norm(centerPs_) #p轨道系数组成向量的长度
    aroundL,aroundL_=np.linalg.norm(aroundPs),np.linalg.norm(aroundPs_)
    # centerRatio=np.linalg.norm(centerPs_)/np.linalg.norm(centerPs) # 投影后与投影前的比例
    # aroundRatio=np.linalg.norm(aroundPs_)/np.linalg.norm(aroundPs) 
    centerRatio=np.divide(centerL_,centerL,out=np.zeros_like(centerL),where=centerL!=0)
    aroundRatio=np.divide(aroundL_,aroundL,out=np.zeros_like(aroundL),where=aroundL!=0)

    
    centerScont=centerAtom.get_sCont(orbital)
    aroundScont=aroundAtom.get_sCont(orbital)
    if centerScont>0.001 or aroundScont>0.001:
        return 0 # s贡献太大的不是

    if centerRatio<=0.1 or aroundRatio<=0.1:
        return 0
    if vector_angle(centerPs_,aroundPs_)<=0.5:
        return 1
    else:
        return -1

def printOrders(orders,orbitals):
    """将轨道根据大小排序后输出"""
    orders_=[f'{order:.4f}' for order in orders]
    sortedRes=sorted(list(zip(orbitals,orders_)),key=lambda e:abs(float(e[1])),reverse=True)
    sortedRes=[e for e in sortedRes if abs(float(e[1]))>=0.01]
    sortedRes=list(zip(*sortedRes))    
    formPrint(sortedRes,8,10)

def formPrint(contents:list[list[str]],eachLength:int,lineNum:int=10):
    """
    格式化打印列表内容，contents是一个列表，其中的每一项是一个包含字符串的列表，每个字符串列表长度必须相同
    contents: 待打印的内容，可以想象成一个表
    eachLength: 打印一行中每一项的字符长度
    lineNum: 一行中有多少项
    """
    logs:list[list[str]]=[]
    for content in contents:
        logs.append([])
        for i in range(0,len(content),lineNum):
            text=''.join([f'{each}'.rjust(eachLength,' ') for each in content[i:i+lineNum]])
            logs[-1].append(text)
    if len(logs)==0:
        return
    for i in range(len(logs[0])):
        for log in logs:
            printer.warn(log[i])




from functools import lru_cache

rawData=[
    [ 1, 'H'],[ 2,'He'],[ 3,'Li'],[ 4,'Be'],[ 5, 'B'],
    [ 6, 'C'],[ 7, 'N'],[ 8, 'O'],[ 9, 'F'],[10,'Ne'],
    [11,'Na'],[12,'Mg'],[13,'Al'],[14,'Si'],[15, 'P'],
    [16, 'S'],[17,'Cl'],[18,'Ar'],[19, 'K'],[20,'Ca'],
    [21,'Sc'],[22,'Ti'],[23, 'V'],[24,'Cr'],[25,'Mn'],
    [26,'Fe'],[27,'Co'],[28,'Ni'],[29,'Cu'],[30,'Zn'],
    [31,'Ga'],[32,'Ge'],[33,'As'],[34,'Se'],[35,'Br'],
    [36,'Kr'],[37,'Rb'],[38,'Sr'],[39, 'Y'],[40,'Zr'],
    [41,'Nb'],[42,'Mo'],[43,'Tc'],[44,'Ru'],[45,'Rh'],
    [46,'Pd'],[47,'Ag'],[48,'Cd'],[49,'In'],[50,'Sn'],
    [51,'Sb'],[52,'Te'],[53, 'I'],[54,'Xe'],[55,'Cs'],
    [56,'Ba'],[57,'La'],[58,'Ce'],[59,'Pr'],[60,'Nd'],
]

class Element:
    def __init__(self,each:list):
        idx,symbol=each
        self.idx=int(idx)
        self.charge=self.idx
        self.symbol=symbol
        self.radius=1.889*0.8

# 因为全局只需要一个Elements实例，因此可以使用类方法
class Elements:
    def __init__(self) -> None:
        self.elements:list[Element]=[Element(each) for each in rawData] #原子列表，一定是按照顺序排列的
        self.symbols:list[str]=[e.symbol for e in self.elements]
        self.charges:list[str]=[e.charge for e in self.elements]
        self.i2e={}
        self.s2e={}

    def __getitem__(self,key:int|str) -> Element:
        if isinstance(key,int):
            assert key in self.charges,f'没有{key}对应元素'
            if key not in self.i2e.keys():
                for element in self.elements:
                    if element.idx==key:
                        self.i2e[key]=element
            return self.i2e[key]
        elif isinstance(key,str):
            assert key in self.symbols,f'没有{key}对应元素'
            if key not in self.s2e.keys():
                for element in self.elements:
                    if element.symbol==key:
                        self.s2e[key]=element
            return self.s2e[key]
        else:
            raise ValueError(f'key必须为str或int类型')

elements=Elements()


"""
模仿multiwfn以命令窗口的形式操作程序计算各种想要的东西
可以彩色输出
绿色-结果
黄色-警告
红色-错误
在没有特别提示的地方,空输入代表返回上级
功能模块划分与pywfn设计一致
"""

from pathlib import Path
import re, sys

from pywfn.reader import get_reader
from pywfn import tools
from pywfn import config
from pywfn.utils import printer
from pywfn import data
from pywfn.base import Mol
import numpy as np


class Shell:
    def __init__(self):
        self.paths: list[Path] = []
        printer.ifShell = True
        printer.info(data.start)
        self.input = Input(self)

    def homePage(self):
        opts = {
            "1": "键の属性",
            "2": "原子属性",
            "3": "实用工具",
            "4": "导出文件",
            "5": "程序设置",
        }
        while True:
            opt = self.input.Option("首页", opts, must=True)
            match opt:
                case "1":
                    self.bondProp()
                case "2":
                    self.atomProp()
                case "3":
                    self.toolsPage()
                case "4":
                    self.writePage()
                case "5":
                    self.setingPage()
                case _:
                    printer.warn("命令不存在")

    def atomProp(self):
        """计算原子属性"""
        from pywfn.atomprop import shell
        shell()
        opts = {
            "1": "电荷分布",
            "2": "电子自旋",
            "3": "π 电子性质",
            "4": "原子自由价",
            "5": "方向电子性质",
            "6": "福井/parr函数",
        }
        chrgs={'1':'mulliken','2':'lowdin'}
        props={'1':'charge','2':'spin'}
        funcs={'1':'fukui','2':'parr'}
        while True:
            opt = self.input.Option("计算原子属性", opts)
            match opt:
                case None:
                    break
                case "1":  # 电荷分布
                    from pywfn.atomprop import charge
                    chrg = self.input.Option(title="输入电荷类型", opts=chrgs, must=True)
                    mol = self.input.Moles()[0]
                    caler = charge.Calculator(mol)
                    if chrg == "1":
                        caler.chrg = "mulliken"
                    if chrg == "2":
                        caler.chrg = "lowdin"
                    caler.logRes()

                case "2":  # 电子自旋
                    chrgs = [["1", "mulliken"], ["2", "lowdin"]]
                    chrg = self.input.Option(title="输入自旋类型", opts=chrgs, must=True)
                    from pywfn.atomprop import spin

                    mol = self.input.Moles()[0]
                    caler = spin.Calculator(mol)
                    if chrg == "1":
                        caler.chrg = "mulliken"
                    if chrg == "2":
                        caler.chrg = "lowdin"
                    caler.logRes()

                case "3":  # π 电子性质
                    from pywfn.atomprop import piProps

                    mol = self.input.Moles()[0]
                    caler = piProps.Calculator(mol)
                    caler.chrg = self.input.Option(title="电荷类型", opts=chrgs, must=True)
                    caler.prop = self.input.Option(title='计算性质', opts=props, must=True)
                    caler.print(caler.resStr())

                case "4":  # 原子自由价
                    from pywfn.atomprop import freeValence

                    mol = self.input.Moles()[0]
                    caler = freeValence.Calculator(mol)
                    while True:
                        idx = self.input.Number(
                            tip="输入原子编号: ", type_="int", length=1
                        )
                        if idx is None:
                            break
                        caler.vect = self.input.Vector(tip="输入法向量", norm=True)
                        caler.logRes()

                case "5":  # 方向电子性质
                    from pywfn.atomprop import dirProps
                    
                    mol = self.input.Moles()[0]
                    caler = dirProps.Calculator(mol)
                    caler.chrg = self.input.Option(title="电荷类型", opts=chrgs, must=True)
                    caler.prop = self.input.Option(title='计算性质', opts=props, must=True)
                    
                    
                    while True:
                        atom = self.input.Number(
                            tip="输入原子编号: ", type_="int", length=1
                        )
                        if atom is None:
                            break
                        vect = self.input.Vector(tip="输入方向向量")
                        if vect is None:
                            break
                        caler.atoms = [atom]
                        caler.vects = [vect]
                        caler.printRes()

                case "6":  # 福井函数、parr函数
                    from pywfn.atomprop import delProps

                    printer.info("分别输入-,0,+对应的分子编号：")
                    mols = self.input.Moles()
                    caler = delProps.Calculator(mols)
                    caler.chrg = self.input.Option(title="电荷类型", opts=chrgs, must=True)
                    caler.prop = self.input.Option(title='计算性质', opts=props, must=True)
                    caler.func = self.input.Option(title='使用函数', opts=funcs, must=True)
                    while True:
                        atom = self.input.Number(
                            tip="输入原子编号: ", type_="int", length=1
                        )
                        if atom is None:
                            break
                        vect = self.input.Vector(tip="输入方向向量：", norm=True)
                        if vect is None:
                            break
                        caler.atoms = [atom]
                        caler.vects = [vect]
                        ev,nv=caler.calculate().flatten() # [1,2]
                        printer.res(f"方向亲电/亲核指标：{ev:.4f},{nv:.4f}")
                        
                case "7":  # 获取各种方向
                    while True:
                        idx = self.input.Number(
                            tip="输入原子编号: ", type_="int", length=1
                        )
                        if idx is None:
                            break
                        atom = mol.atom(idx)

                        normal = atom.get_Normal()
                        obtWay = atom.get_obtWay(len(mol.O_obts) - 1)

                        if normal is not None:
                            printer.vector("原子法向量", normal)
                        else:
                            printer.warn("原子没有法向量")
                        printer.vector("HOMO轨道方向", obtWay)
                case _:
                    return

    def bondProp(self):
        """计算各种键级"""
        opts = {
            "1": "π 键级(OP)",
            "2": "π 键级(SMO)",
            "3": "Mayer 键级",
            "4": "HMO 键级",
        }
        mol=self.input.Moles()[0]
        while True:
            match self.input.Option("计算各种键级", opts):
                case None:
                    break
                case "1":
                    from pywfn.bondprop import piDM

                    caler = piDM.Calculator(mol)
                    while True:
                        bond = self.input.Number(
                            "输入两原子的编号(,)[]: ", type_="int", length=2
                        )
                        if bond is None:
                            break
                        caler.bond = bond
                        caler.print()
                case "2":
                    from pywfn.bondprop import piSM

                    caler = piSM.Calculator(mol)
                    while True:
                        bond = self.input.Number(
                            "输入两原子的编号(,)[]: ", type_="int", length=2
                        )
                        if bond is None:
                            break
                        caler.bond = bond
                        caler.print()
                case "3":
                    from pywfn.bondprop import mayer

                    caler = mayer.Calculator(mol)
                    while True:
                        bond = self.input.Number(
                            "输入两原子的编号(,)[]: ", type_="int", length=2
                        )
                        if bond is None:
                            break
                        caler.bond = bond
                        caler.print()
                case "4":
                    from pywfn.bondprop import hmo

                    caler = hmo.Calculator(mol)
                    while True:
                        bond = self.input.Number(
                            "输入两原子的编号(,)[]: ", type_="int", length=2
                        )
                        if bond is None:
                            break
                        caler.bond = bond
                        caler.print()
                case _:
                    printer.warn("选项不存在")
                    continue

    def toolsPage(self):
        """进入实用工具页面"""
        opts = {
            "1": "分割 扫描 文件",
            "2": "分割 IRC 文件",
            "3": "分割 link 任务",
            "4": "生成 PES 文件",
            "5": "计算 反应活化能",
            "6": "计算 立体选择性ee值",
        }
        while True:
            opt = self.input.Option("使用工具页面", opts)
            match opt:
                case None:
                    break
                case "1":
                    for i, path in enumerate(self.paths):
                        tools.SplitScan(path).split()
                    printer.res("文件分割完成 >_<")
                case "2":
                    for i, path in enumerate(self.paths):
                        tools.SplitIrc(path).split()
                    printer.res("文件分割完成 >_<")
                case "3":
                    for i, path in enumerate(self.paths):
                        tools.SplitLink(path).split()
                case "4":
                    from pywfn.tools.getPES import Tool

                    self.showFiles()
                    tool = Tool(self.paths)
                    tool.route = self.input.Number(tip="路径索引：", type_="int")
                    tool.create()
                case "5":
                    from pywfn.chemProp import reaActEne
                    K = self.input.Number(tip="输入K(催化常数): ", length=1)
                    T = self.input.Number(tip="输入T(反应温度,K): ", length=1)
                    g0 = reaActEne(K,T)
                    g1 = g0 / 4.184
                    printer.res(
                        f"G(吉布斯自由能):\n{g0:>10.4f} kJ/mol\n{g1:>10.4f} Kcal/mol"
                    )
                case "6":
                    from pywfn.chemProp import steSelEE
                    deR, deS = self.input.Number(tip="分别输入ΔG(R),ΔG(S): ", length=2)
                    T = self.input.Number(tip="输入T(反应温度,K): ", length=1)
                    ee = steSelEE(deR,deS,T)
                    printer.res(f"立体选择性ee值:\n{ee*100:.2f}%")

    def writePage(self):
        while True:
            opts = {
                "1": "gif 文件",
                "2": "xyz 文件",
                "3": "cub 文件",
                "4": "si  文件",
            }
            opt = self.input.Option("导出文件", opts)
            match opt:
                case None:
                    return
                case "1":
                    from pywfn.writer import gjfWriter

                    for path in self.paths:
                        mol = Mol(get_reader(path))
                        gjfWriter(mol).save()
                case "2":
                    from pywfn.writer import xyzWriter

                    for path in self.paths:
                        mol = Mol(get_reader(path))
                        xyzWriter(mol).save()
                case "3":
                    from pywfn.writer import cubWriter

                    obts = self.input.Number(tip="请输入轨道序号(,)[]: ", type_="int")
                    obts = [obt - 1 for obt in obts]
                    step = self.input.Number(
                        tip=f"请输入渲染间隔()[{config.RENDER_CLOUD_STEP}]: ",
                        type_="float",
                        length=1,
                    )
                    atoms = self.input.Number(
                        tip="请输入渲染原子(,)[全选]: ", type_="int"
                    )
                    direct = self.input.Vector(
                        tip="请输入投影方向[]", norm=True, must=False
                    )
                    for path in self.paths:
                        mol = Mol(reader=get_reader(path))
                        writer = cubWriter(mol)
                        if step is not None:
                            writer.step = step
                        if atoms is not None:
                            writer.atoms = atoms
                        if direct is not None:
                            writer.direct = direct
                        writer.obts = obts
                        writer.save("")
                case "4":
                    from pywfn.writer import siWriter

                    printer.info("1.坐标 2.能量 3.频率")
                    selects = self.input.Number(tip="选择导出信息[全选]: ")
                    same = self.input.Bool(tip="保存同一文件?([y]/n): ")
                    for path in printer.track(self.paths, tip="正在导出"):
                        reader = get_reader(path)
                        writer = siWriter(Mol(reader=reader))
                        if selects is not None:
                            writer.selects = selects
                        writer.sameFile = same
                        writer.save()

    def showFiles(self):
        printer.info(f"当前读取{len(self.paths)}文件：")
        for i, path in enumerate(self.paths):
            print(f"{i},{path.stem}")
        printer.print("")

    def setingPage(self):
        """设置页面"""
        from pywfn.data import temps

        opts = {
            "1": "gif 模板",
            "2": "si  模板",
        }
        while True:
            opt = self.input.Option("设置页面", opts)
            match opt:
                case None:
                    return
                case "1":
                    path, text = self.input.Text()
                    temps.gjf = text
                    config.set_config("TEMPLATE_PATH_GJF", path)
                case "2":
                    path, text = self.input.Text()
                    temps.si = text
                    config.set_config("TEMPLATE_PATH_SI", path)
            printer.res("设置成功")


class Input:
    def __init__(self, shell: Shell) -> None:
        self.shell = shell

    def input(self, tip: str):
        res = input(tip)
        if res == "q":
            sys.exit()
        return res

    def Number(self, tip: str, type_: str = "float", length: int = None) -> int | float:
        """输入整数"""

        def split1(frg: str):
            if "-" in frg:
                num1, num2 = [int(e) for e in frg.split("-")]
                assert num1 < num2, "数字1要大于数字2"
                return list(f"{e}" for e in range(num1, num2 + 1))
            else:
                return [frg]

        while True:
            opt = input(tip)
            if opt == "":
                return None
            else:
                if not re.match(r"^-?\d+[\.\d+]*([,，-]-?\d+[\.\d+]*)*$", opt):
                    printer.warn("格式不正确！！")
                    continue

                frgs = re.split(r"[,，]", opt)
                nums = []
                for frg in frgs:
                    nums += split1(frg)
                nums = [int(num) if type_ == "int" else float(num) for num in nums]
                if length is None:
                    return nums
                else:
                    if len(nums) != length:
                        continue
                    if length == 1:
                        return nums[0]
                    else:
                        return nums

    def Vector(self, tip: str, norm: bool = False, must=False) -> np.ndarray:
        """
        输入向量
        tip:提示
        norm:是否归一化
        must:是否必须
        """
        while True:
            opt = input(tip)
            if opt == "":
                if must:
                    continue
                else:
                    return None
            nums = re.split(r"[,，]", opt)
            if len(nums) != 3:
                continue
            nums = np.array(nums, dtype=np.float32)
            if norm:
                nums /= np.linalg.norm(nums)
            return nums

    def Bool(self, tip: str, default: bool = True):
        opt = input(tip)
        match opt:
            case "":
                return default
            case "y":
                return True
            case "n":
                return False
            case _:
                printer.warn("请输入正确选项!")
                return self.Bool(tip, default)

    def Option(self, title: str, opts:dict[str,str], must=False):
        # maxLen=max([len(text) for idx,text in opts])
        printer.options(title, opts)
        idxs = opts.keys()
        while True:
            opt = self.input("请输入对应选项: ")
            if opt == "":
                if must:
                    continue
                else:
                    return None
            elif opt in idxs:
                return opt
            elif Path(opt).exists():  # 如果输入的是一个文件
                self.Files(opt)
            elif opt == "sf":
                self.shell.showFiles()
            elif opt == "cf":
                printer.info("清空文件")
                self.shell.paths = []
                self.shell.showFiles()
            else:
                printer.warn("无效的输入!")

    def Files(self, path:str):
        types = [".log", ".out", ".fch", ".gjf"]  # 支持的文件类型
        pathObj = Path(path)
        if pathObj.is_file():  # 如果是文件
            if pathObj.suffix in types:
                printer.info(f"共1个文件")
                self.shell.paths += [path]
            else:
                printer.warn("不支持的文件类型")
        elif pathObj.is_dir():  # 如果是文件夹
            paths = []
            for each in pathObj.iterdir():  # 对文件夹中的每个文件进行循环
                if each.suffix in types:  # 如果文件类型是支持的文件类型
                    paths.append(each)
            printer.info(f"输入{len(paths)}个文件")
            self.shell.paths += paths

    def Text(self) -> str:
        path = input("请输入文件: ")
        if not Path(path).exists():
            return None
        return path, Path(path).read_text()

    def Moles(self) -> list[Mol]:
        paths = self.shell.paths
        if len(paths) == 1:
            path = paths[0]
            reader = get_reader(path)
            return [Mol(reader=reader)]
        self.shell.showFiles()
        idxs = self.Number("输入分子编号：", type_="int")

        mols = []
        for idx in idxs:
            path = paths[idx]
            reader = get_reader(path)
            mol = Mol(reader=reader)
            mols.append(mol)
        return mols



import numpy as np

from pywfn import base

def pola2cart(r,t,p):
    """极坐标转直角坐标，半径，仰角，转角"""
    z=r*np.sin(t)
    x=np.cos(t)*np.cos(p)
    y=np.cos(t)*np.sin(p)
    return x,y,z

def get_direct(atom:"base.Atom"):
    """
    根据最大排斥原则，迭代寻找反应方向
    在原子的外层采点，计算目标值，目标值最小作为初猜
    """
    def get_score(point:np.ndarray): # 衡量当前点好坏的一个函数
        point=point.reshape(1,3)
        angles=-np.dot(point,vectors)+1 # 不是真正的角度，但与真正的角度正相关
        mean=np.mean(angles)
        score=sum([ang**2-(ang-mean)**2 for ang in angles])
        return score

    points=np.zeros(shape=(50,3))
    for t in np.linspace(0,np.pi,5):
        for p in np.linalg(0,2*np.pi,10):
            x,y,z=pola2cart(1,t,p)
            points[t*5+p]=[x,y,z]

    vectors=np.concatenate([around.coord-atom.coord for around in atom.neighbors],axis=0)
    
    scores=np.array([get_score(point) for point in points])
    maxScore=np.max(scores)
    index=np.argmax(scores)
    point0=points[index] # 获取初猜
    arounds=np.array([
        [-1,0],
        [ 1,0],
        [0, 1],
        [0,-1]
    ])
    stepSize=np.pi/10
    for i in range(100):
        points=point0+arounds*stepSize
        scores=np.array([get_score(point) for point in points])
        maxScore_=np.max(scores)
        index=np.argmax(scores)
        point0_=points[index]
        if maxScore_-maxScore>1e-5:
            point0=point0_
            maxScore=maxScore_
        else:
            if stepSize<np.pi/100:
                return point0
            else:
                stepSize/=2


from pywfn.utils import printer
class Caler:

    def calculate(self,idx1:int,idx2:int)->float:
        return 0.0

    def print(self,resStr:str):
        printer.res(resStr)

    def resStr(self,idx1:int,idx2:int)->str:
        res=self.calculate(idx1,idx2)
        return f'{res}'


"""
需要多个分子才能计算的属性
"""

from functools import cached_property,lru_cache
import numpy as np

from pywfn import base
from pywfn.atomprop import charge
from pywfn.molprop import MolProp
from pywfn.utils import printer

class Fukui:
    """
    福井函数，以及依赖于福井函数计算的一些性质
    福井函数值针对原子的函数
    """
    def __init__(self,molN:"base.Mol",molM:"base.Mol",molP:"base.Mol") -> None:
        """
        需要三个分子对象
        molN:负电性分子
        molM:中性分子
        molP:正电性分子
        """
        
        if len(set([len(molN.atoms),len(molM.atoms),len(molP.atoms)]))!=1:
            printer.warn('三个分子的原子数量不相等')
        calerN=charge.Calculator(molN)
        calerM=charge.Calculator(molM)
        calerP=charge.Calculator(molP)
        chargesN=calerN.calculate(atoms=None)
        chargesM=calerM.calculate(atoms=None)
        chargesP=calerP.calculate(atoms=None)
        self.fns=np.array(chargesM)-np.array(chargesN)
        self.fps=np.array(chargesP)-np.array(chargesM)
        self.mol=molM
        self.molProp=MolProp(molM)
    
    @lru_cache
    def fn(self,idx:int):
        """获取某个原子的福井函数+"""
        return self.fns[idx-1]
    
    @lru_cache
    def fp(self,idx:int):
        """获取某个原子的福井函数-"""
        return self.fps[idx-1]
    
    @lru_cache
    def cs(self,idx:int):
        """局部亲核指标"""
        cs=self.molProp.cs
        return cs*self.fn(idx),cs*self.fp(idx)
    
    @lru_cache
    def ei(self,idx:int):
        """局部亲电指标"""
        ei=self.molProp.ei
        return ei*self.fn(idx),ei*self.fp(idx)


"""
用来生成格点数据并保存cube文件
cube文件使用的是玻尔半径B
分子的坐标使用的是埃米A
A=B/1,889
B=A*1.889
计算的时候使用A，写文件的时候将坐标和格点转为B
"""

import numpy as np
import time
from tqdm import tqdm
from pathlib import Path

from pywfn import maths,base,config
from pywfn.utils import printer
from pywfn.writer import lutils


class cubWriter:
    def __init__(self,mol:base.Mol) -> None:
        self.mol=mol
        self.step:float=config.RENDER_CLOUD_STEP
        self.border:float=config.RENDER_CLOUD_BORDER
        self.atoms:list[int]=[atom.idx for atom in self.mol.atoms]
        self.direct:np.ndarray=None
        self.floatNum=12
        self.obts=[]
    
    def init_file(self,name:str):
        self.title0='genetrate by pywfn'
        self.title1=time.strftime('%Y-%m-%d %H:%M:%S')
        path=Path(self.mol.reader.path)
        # obts=','.join((f'{e+1}' for e in self.obts))
        self.filePath=(path.parent/f'{path.stem}_{name}.cub')
        self.file=open(self.filePath,mode='w')
        

    def get_gridPos(self):
        """生成格点数据"""
        self.mol.bohr=False
        atoms=[atom-1 for atom in self.atoms]
        p0=self.mol.coords[atoms,:].min(axis=0)
        p1=self.mol.coords[atoms,:].max(axis=0)
        bord=self.border
        
        (Nx,Ny,Nz),gridPos=maths.gridPos(p0-bord,p1+bord,self.step,getN=True) #计算波函数时的坐标还是要使用原子单位的坐标
        self.gridSize=(Nx,Ny,Nz)
        assert Nx*Ny*Nz==len(gridPos),"网格点数量不匹配"
        self.file.write(f'{self.title0}\n{self.title1} {len(gridPos)*len(self.obts)}\n')
        x0,y0,z0=[e*config.BOHR_RADIUS for e in p0-bord]
        step=self.step*config.BOHR_RADIUS
        natm=len(self.mol.atoms)
        self.file.write(f'{-natm:>5}{x0:>12.6f}{y0:>12.6f}{z0:>12.6f}\n')
        self.file.write(f'{Nx:>5}{step:>12.6f}{0:>12.6f}{0:>12.6f}\n')
        self.file.write(f'{Ny:>5}{0:>12.6f}{step:>12.6f}{0:>12.6f}\n')
        self.file.write(f'{Nz:>5}{0:>12.6f}{0:>12.6f}{step:>12.6f}\n')
        self.write_coord()
        return gridPos
    

    
    def write_coord(self):
        
        for atom in self.mol.atoms:
            x,y,z=atom.coord*config.BOHR_RADIUS
            self.file.write(f'{atom.atomic:>5}{atom.atomic:12.6f}{x:12.6f}{y:12.6f}{z:12.6f}\n')
    
    def write_value(self):
        obts=self.obts
        gridPos=self.get_gridPos()
        Nx,Ny,Nz=self.gridSize
        obtInfo=[len(obts)]+[o+1 for o in obts]

        for i,info in enumerate(obtInfo): # 写入轨道信息
            self.file.write(f'{info:>5}')
            if(i+1)%10==0:self.file.write('\n')
        if(i+1)%10!=0:self.file.write('\n')
        allVals=[self.mol.get_wfnv(obt,gridPos,self.atoms) for obt in obts]
        lenVals=len(gridPos)
        index=0
        for i in range(lenVals):
            for j in range(len(allVals)):
                v=allVals[j][i]
                # if v==0:v=1e-8
                self.file.write(f'{v:13.5E}')
                index+=1
                if index==Nz*len(obts):
                    self.file.write('\n')
                    index=0
                    continue
                if index%6==0:self.file.write('\n')
    
    def save(self,name):
        if self.direct is not None:
            self.mol.projCM(self.atoms,self.obts,[self.direct.copy()]*len(self.atoms),True,False)
        self.init_file(name)
        self.write_value()
        printer.res(f'导出文件至{self.filePath}')
        self.file.close()


"""
将分子导出为xyz文件
"""
from pywfn import base
from pathlib import Path
from pywfn.data.elements import elements
from pywfn.utils import printer

class xyzWriter:
    def __init__(self,mol:"base.Mol") -> None:
        self.mol=mol
        self.resStr=''

    def write(self):
        self.resStr+=f'{len(self.mol.atoms)}\n'
        self.resStr+=f'generate by pywfn\n'
        for atom in self.mol.atoms:
            x,y,z=atom.coord
            sym=atom.symbol
            idx=elements[sym].charge
            self.resStr+=f' {idx:<14}{x:>14.8f}{y:>14.8f}{z:>14.8f}\n'

    def save(self):
        self.write()
        path=self.mol.reader.path
        path=Path(path)
        filePath=(path.parent/f'{path.stem}.xyz')
        filePath.write_text(self.resStr)
        print(filePath)
        printer.res(f'{filePath} 导出成功!😄')


"""
计算一些简单的物理化学性质，不需要输入文件
"""
import numpy as np
from pywfn import consts as cs
# 计算反应活化能 reaction activation energy
def reaActEne(K:float,T:float)->float:
    """
    计算化学反应活化能,KJ/mol
    K:催化常数
    T:反应温度 开尔文
    """
    c = 2 * 1e10 * T
    x = np.log(K / c)
    g = -x * cs.R * T * 1e-3  # kJ/mol
    # g1 = g / 4.184
    return g

# 计算立体选择ee值 Stereo selection ee value
def steSelEE(deR,deS,T):
    RT = cs.R * T
    RS = np.exp((deS - deR) * 4185.8518 / RT)
    ee = (RS - 1) / (RS + 1)
    return ee


"""
分割IRC文件，IRC文件没有每一步的结构优化，所以只需要提取坐标即可
"""
import re,os
from pathlib import Path

from pywfn.tools.fileCreater import Tool as  FileCreater
from pywfn.data import Elements
from pywfn.utils import printer

class Tool:
    def __init__(self,path:str) -> None:
        self.path=Path(path)
        self.dirName=self.path.parent
        self.baseName=self.path.name # 包含后缀的文件名
        self.fileName=self.path.stem # 不包含后缀的文件名
        folder=self.dirName / self.fileName
        if not Path(folder).exists(): #判断文件夹是否存在
            os.mkdir(folder)
        self.content=self.path.read_text(encoding='utf-8')
        self.lines=self.content.splitlines(keepends=False)
        if 'Standard orientation' in self.content:
            self.coordType='Standard orientation'
        else:
            self.coordType='Input orientation'
        self.elements=Elements()
    
    def find_titles(self):
        titleNums=[]
        idx=0
        for i,line in enumerate(self.lines): # 循环每一行
            if self.coordType in line: # 找到行坐标
                titleNums.append((idx,i)) # 只要第一个,记录段落数和行数
                idx+=1
        return titleNums
    
    def match_coords(self,titleNums:list[tuple[int]]):
        s=r" +\d+ +(\d+) +\d+ +(-?\d+.\d+) +(-?\d+.\d+) +(-?\d+.\d+)"
        coords=[]
        for idx,titleNum in titleNums:
            coord=[]
            for i in range(titleNum+5,len(self.lines)):
                line=self.lines[i]
                if re.search(s,line) is not None:
                    idx,x,y,z=re.search(s,line).groups()
                    symbol=self.elements.get_element_by_idx(int(idx)).symbol
                    coord.append([symbol,x,y,z])
                else:
                    break
            coords.append(coord)
        return coords

    def split(self):
        """分割文件"""
        # 首先获取所有构象的坐标
        titleNums=self.find_titles()
        coords=self.match_coords(titleNums)
        for i,coord in printer.track(enumerate(coords)):
            path=os.path.join(self.dirName,self.fileName,f'f{i+1:0>2}.gjf')
            fileCreater=FileCreater(path=path)
            fileCreater.set_coord(coord)
            fileCreater.set_chk(f'{i+1}')
            fileCreater.save()


import numpy as np
import math
from pywfn import config
from pywfn import base
from pywfn.utils import printer
from pywfn.maths.gto import Gto


def gridPos(
    p0: np.ndarray, p1: np.ndarray, step: float, bord: float = 0, getN: bool = False
):
    """生成网格数据点,range:生成数据的范围
    getN:是否获取每个维度的数量
    getR:是否获取每个维度的长度
    """
    assert isinstance(p0, np.ndarray), "必须是np.ndarray类型"
    x0, y0, z0 = p0 - bord
    x1, y1, z1 = p1 + bord
    pos = []
    xs = np.arange(x0, x1, step)
    ys = np.arange(y0, y1, step)
    zs = np.arange(z0, z1, step)
    Nx, Ny, Nz = len(xs), len(ys), len(zs)
    for x in xs:
        for y in ys:
            for z in zs:
                pos.append([x, y, z])

    if getN:
        return (Nx, Ny, Nz), np.array(pos, dtype=np.float32)
    else:
        return np.array(pos, dtype=np.float32)


def CM2PM(CM, obts: list[int], oe: int) -> np.ndarray:
    """
    根据系数矩阵构建密度矩阵
    CM:系数矩阵,如果是开壳层的话,列数是行数的两倍[n,n]/[n,2n]
    n:分子轨道占据电子数
    """
    CMv=CM[:,obts]
    CMh=np.transpose(CMv)
    PMs = CM2PMs(CM, obts, oe)
    return np.sum(PMs, axis=0)


def CM2PMs(CM, obts: list[int], oe: int):
    """
    构建三维密度矩阵，不要空轨道，形状为[占据轨道数,原子轨道数,原子轨道数]
    """
    A = (CM[:, obts].T)[:, :, np.newaxis]
    B = (CM[:, obts].T)[:, np.newaxis, :]
    return A @ B * oe  # 用矩阵乘法的形式直接构建矩阵可比逐元素计算多了


def vector_angle(a: np.ndarray, b: np.ndarray) -> float:  # 计算两向量之间的夹角
    """
    计算两向量之间的夹角
    a:o->a
    b:o->b
    acute:是否为钝角

    """
    la = np.linalg.norm(a)
    lb = np.linalg.norm(b)
    assert la * lb > 0, "向量长度不应为0"

    value = np.dot(a, b) / (la * lb)  # 应介于(-1,1)之间
    if value > 1:
        value = 1
    if value < -1:
        value = -1
    angle = np.arccos(value) / np.pi
    return angle


def get_normalVector(p1, p2, p3, linear=False):
    """
    获取三点确定的平面的单位法向量
    根据两个向量也可以确定法向量
    """
    vi = p3 - p2
    vj = p1 - p2
    n = np.cross(vi, vj)  # 法向量
    if (
        np.linalg.norm(n) == 0
    ):  # 此时说明三个原子在一条直线上，是标准的线型分子，所以不存在法向量
        return None
    if vector_angle(vi, vj, trans=True) < 0.02 and linear:
        return None
    if vector_angle(n, config.BASE_VECTOR) > 0.5:
        n *= -1
    return n / np.linalg.norm(n)  # 返回单位向量


def linear_classify(points):  # 将向量分类转为表示角度的数值分类
    nv = points[-1]
    angles = np.array([vector_angle(each, nv) for each in points])
    y = np.abs(0.5 - angles)
    types = [[], []]
    idxs = [[], []]
    for i, each in enumerate(y):
        distance = np.abs(each - np.array([0, 0.5]))
        idx = np.argmin(distance)
        types[idx].append(each)
        idxs[idx].append(i)
    return idxs


def orbital_classify(atom, vectors, O_obts):
    """定义函数将轨道分类"""
    points = []
    orbitals = []
    keys = vectors.keys()
    for key in keys:
        atomidx = int(key.split("-")[0])
        orbitalidx = int(key.split("-")[1])
        if atomidx == atom and orbitalidx in O_obts and vectors[key] is not None:
            vector = vectors[key] / np.linalg.norm(vectors[key])
            points.append(vector)
            orbitals.append(orbitalidx)
    if len(points) > 0:
        points = np.array(points)
        idxs = linear_classify(points)
    else:
        idxs = [[], []]
    V, H = [[orbitals[id] for id in idx] for idx in idxs]
    return V, H


def get_aroundPoints(p, step):  #
    """
    计算空间中一个点周围六个点处的坐标
    p:[3,]坐标
    step:步长
    """
    arounds = np.array(
        [[-1, 0, 0], [+1, 0, 0], [0, -1, 0], [0, +1, 0], [0, 0, -1], [0, 0, +1]]
    )
    np.random.shuffle(arounds)
    return p + arounds * step


def get_extraValue(atom: "base.Atom", obt: int, valueType="max"):
    """
    从指定位置开始,利用爬山算法寻找原子波函数极值
    maxPos:[3,]
    """
    # p0=atom.coord.copy() # 起始点,p是原子坐标不变
    # p=p0.copy()
    p0 = np.array([0.0, 0.0, 0.0]).reshape(1, 3)
    v0 = atom.get_wfnv(p0, obt)  # 计算原子坐标处的初始值
    step = 0.1
    while True:
        aroundPs = get_aroundPoints(p0, step)  # aroundPs:(n,3)
        aroundVs = atom.get_wfnv(aroundPs, obt)
        if valueType == "max" and np.max(aroundVs) > v0:
            maxID = np.argmax(aroundVs)  # 最大值的索引
            p0 = aroundPs[maxID]  # 最大值坐标
            v0 = aroundVs[maxID]  # 最大值
        elif valueType == "min" and np.min(aroundVs) < v0:
            minID = np.argmin(aroundVs)
            p0 = aroundPs[minID]
            v0 = aroundVs[minID]
        else:
            if step <= 1e-6:
                return p0, v0  #
            else:
                step /= 10


def search_sp2(idx: int, mol: "base.Mol") -> int:
    """
    深度优先搜索方法寻找于指定原子相邻最近的sp2 C原子
    """
    atom = mol.atom(idx)

    searchd = [idx]  # 已经搜索过的
    searchs = [a.idx for a in atom.neighbors]  # 将要搜索的

    while len(searchs) > 0:
        idx = searchs.pop(0)  # 弹出第一个
        atom = mol.atom(idx)
        if len(atom.neighbors) == 3:
            return idx
        searchn = [
            a.idx for a in mol.atom(idx).neighbors if a.idx not in searchd
        ]  # 新找到的原子的索引
        searchs += searchn
    printer.warn("没有找到sp2类型原子")
    return None


def points_rotate(
    points: np.ndarray, center: np.ndarray, axis: np.ndarray, angle: float
):  # 旋转点
    """
    将一个点或一系列点绕空间中的某个轴旋转指定角度，输入和输出都是空间中点的绝对位置
    points:要旋转的点
    center:旋转中心
    axis:旋转轴
    angle:旋转角度
    """
    assert len(points.shape) == 2, "旋转点应该为二维数组"
    assert points.shape[1] == 3, "旋转点的形状应为[n,3]"
    axis /= np.linalg.norm(axis)
    nx,ny,nz=axis
    c=np.cos(angle)
    s=np.sin(angle)
    matrix = np.array([
        [nx**2*(1-c)+c,nx*ny*(1-c)+nz*s,nx*nz*(1-c)-ny*s],
        [nx*ny*(1-c)-nz*s,ny**2*(1-c)+c,ny*nz*(1-c)+nx*s],
        [nx*nz*(1-c)+ny*s,ny*nz*(1-c)-nx*s,nz**2*(1-c)+c],
    ])
    length0 = np.linalg.norm(points, axis=1)
    points = np.copy(points)
    points -= center
    points = np.dot(matrix, points.T).T
    points += center
    length1 = np.linalg.norm(points, axis=1)
    return points


def get_plane_by_3points(
    pos1: np.ndarray, pos2: np.ndarray, pos3: np.ndarray
) -> list[float]:
    """
    根据三个点计算平面
    平面方程：ax+by+cz+d=0

    """
    v21 = pos1 - pos2
    v23 = pos3 - pos2
    a, b, c = np.cross(v21, v23)
    x1, y1, z1 = pos1
    d = -a * x1 - b * y1 - c * z1
    return [a, b, c, d]


def get_distance_to_plane(params: list[float], pos: np.ndarray) -> float:
    """
    计算点到平面距离
    params：a,b,c,d
    """
    a, b, c, d = params
    x, y, z = pos
    distan = np.abs(a * x + b * y + c * z + d) / np.sqrt(a**2 + b**2 + c**2)
    return distan



"""
此脚本用来计算mulliken自旋
自旋是α电子-β电子
可以通过改变轨道的占据情况来分别计算α和β电子数
"""

from pywfn.base import Mol,Atom
from pywfn.utils import printer
from pywfn.atomprop import charge, lutils,AtomCaler
from pywfn.atomprop.charge import Chrgs
from pywfn.maths import CM2PM

from typing import Literal
import numpy as np

class Calculator(AtomCaler):
    def __init__(self,mol:"Mol") -> None:
        self.mol=mol
        self.logTip='mulliken 电子自旋分布:'

    def calculate(self,chrg:Chrgs='mulliken')->np.ndarray:
        """计算所有原子的自旋"""
        assert self.mol.open,'非开壳层分子无法计算自旋'
        obtNum=self.mol.CM.shape[0] # 系数矩阵行数，基函数数量
        occs_old=self.mol.obtOccs # 记录原本的占据情况
        a_occs=occs_old.copy() # 当你需要修改一个变量的时候，
        b_occs=occs_old.copy()
        a_occs[obtNum:]=[False]*obtNum
        b_occs[:obtNum]=[False]*obtNum

        caler=charge.Calculator(self.mol)

        # 将长方形的系数矩阵分为两个正方形分别计算

        self.mol.props['obtOccs']=a_occs
        a_Ects=caler.calculate(chrg=chrg)

        self.mol.props['obtOccs']=b_occs
        b_Ects=caler.calculate(chrg=chrg)
        # 恢复分子属性
        self.mol.props['obtOccs']=occs_old
        return -(a_Ects-b_Ects)
        
    def resStr(self):
        elects=self.calculate()
        atoms=lutils.atomIdxs(self.mol.atoms)
        return lutils.atomValueStr(self.mol,atoms,elects)
    
    def onShell(self):
        spins=self.calculate()
        for i,spin in enumerate(spins):
            printer.res(f'{i+1}: {spin}')
        return


"""
此脚本用来读取fchk文件
fchk文件中有哪些属性是可以用到的？
"""
import re
import numpy as np

from pywfn import base
from pywfn.data.elements import elements
from pywfn.reader.lutils import Reader
from collections import defaultdict

class Title:
    def __init__(self,lineNum:int,dataNum:int,dataType:str,hasData:bool) -> None:
        self.lineNum=lineNum
        self.dataNum=dataNum
        self.dType=dataType #数据类型
        self.hasData=hasData # 是否包含更多数据

titleMatch='^(.{40}) {3}(.{1})(.{5})(.{12})$'
class FchReader(Reader):
    def __init__(self,path:str):
        super().__init__(path)
        self.mol=base.Mol()

        self.jobTitle:str=self.lines[0]
        mathc2=re.match(r'(.{10})(.{30})(.{30})',self.lines[1]).groups()
        self.jobType,self.jobMethod,self.jobBasis=mathc2
        self.needs=[
            'Atomic numbers',
            'Current cartesian coordinates',
            'Shell to atom map',
            'Primitive exponents',
            'Contraction coefficients',
            'Alpha Orbital Energies',
            'Alpha MO coefficients',
            'Mulliken Chrgs'
        ]
        self.titles:dict[str,int]={}
        self.marks=[
            [0,40],
            [40,43],
            [43,45],
            [45,50],
            [50,62]
        ]

    def search_title(self):
        """搜索标题行"""
        patern=r'[A-Za-z -/]{40}.{3}[IRc]{1}.{5}[ \d]{12}'
        for l,line in enumerate(self.lines):
            if line[0]==' ':continue
            title=line[0:40].strip()
            if title not in self.needs:continue
            self.titles[title]=l
            
    def parse_title(self,title:str):
        lineNum=self.titles[title]
        line=self.lines[lineNum]
        finds=[line[u:l] for u,l in self.marks]
        _,_,dtype,hasData,dataNum=finds
        dtype=dtype.strip()
        hasData=bool(hasData)
        dataNum=int(dataNum)
        lineSpan=dataNum//6+(0 if dataNum%6==0 else 1)
        tpmap={'I':'\d+','R':'-?\d+.\d+'}
        text='\n'.join(self.lines[lineNum+1:lineNum+1+lineSpan])
        values=re.findall(tpmap[dtype],text)
        if dtype=='I':
            values=[int(v) for v in values]
        if dtype=='R':
            values=[float(v) for v in values]
        return values
    
    def get_symbols(self) -> list[str]:
        atomics=self.read_atoms()
        symbols=[elements[a].symbol for a in atomics]
        return symbols

    def read_atoms(self):
        values=self.parse_title('Atomic numbers')
        print(values)
        return values

    def get_coords(self) -> np.ndarray:
        coords=self.read_coord()
        return coords

    def read_coord(self):
        values=self.parse_title('Current cartesian coordinates')
        values=np.array(values,dtype=np.float32).reshape(-1,3)/1.889726
        print(values)
        return values



from pywfn.base import Mol
from pywfn.maths import vector_angle

import numpy as np

def dihedralAngle(mol:Mol,idxs:list[int]):
    """计算二面角"""
    a,b,c,d=[mol.atom(idx) for idx in idxs]
    vba=a.coord-b.coord
    vbc=c.coord-b.coord
    vcd=d.coord-c.coord
    vcb=b.coord-c.coord
    vi=np.cross(vba,vbc)
    vj=np.cross(vcb,vcd)
    angle=vector_angle(vi,vj)
    return angle


"""
计算分子轨道内每个原子的能量
"""
from pywfn.base import Mol
from pywfn.molprop import obtEnergy
from pywfn.atomprop import lutils
from pywfn.utils import printer
import numpy as np

class Calculator:
    def __init__(self,mol:Mol) -> None:
        self.mol=mol
        self.caler=obtEnergy.Calculator(mol)
    
    def calculate(self)->np.ndarray:
        EM=self.caler.calculate() # 获取能量矩阵
        atoms=self.mol.atoms
        engs=np.zeros(len(atoms))
        for a,atom in enumerate(self.mol.atoms):
            u,l=atom.obtBorder
            atomEng=EM[u:l,:].sum()
            engs[a]=atomEng
        return engs
        
    def printRes(self):
        resStr=self.resStr()
        printer.info('原子能量分布: ')
        printer.res(resStr)
        
    def resStr(self):
        engs=self.calculate()
        atoms=[a.idx for a in self.mol.atoms]
        return lutils.atomValueStr(self.mol,atoms,engs)
    
    def onShell(self):
        engs=self.calculate()
        for i,eng in enumerate(engs):
            printer.res(f'{i}: {eng}')
        return


"""
该子模块定义各种文件的读取器
所有读取器应该都是读取分子所需要的属性？
数据尽量向量化存储
分子文件需要哪些属性，则读取器需要读取哪些属性
原子类型
原子坐标
系数矩阵
重叠矩阵
"""
from pywfn.reader.lutils import Reader
from pywfn.reader.fch import FchReader
from pywfn.reader.log import LogReader
from pywfn.reader.gjf import GjfReader
from pywfn.reader.mol import MolReader

from pathlib import Path


# 定义所有reader的基类
# 基类中定义各种属性的读取函数，再在子类中覆盖



def get_reader(path:str):
    """根据输入文件的类型自动判断应该使用哪个读取器"""
    suffix=Path(path).suffix
    readers={
        '.out':LogReader,
        '.log':LogReader,
        '.gjf':GjfReader
    }
    assert suffix in readers.keys(),'不支持的读取'
    return readers[suffix](path)
    


"""
各种带方向的原子属性放到一块吧
"""
from pywfn.base import Mol
from pywfn.atomprop import charge,freeValence, spin
from pywfn.atomprop import energy
from pywfn.atomprop import lutils
from pywfn.utils import printer
import numpy as np
from typing import Literal


class Calculator:
    """
    带有方向的各种原子属性的计算
    需要指定：电荷(chrg),prop(性质),atoms(原子),vects(方向)
    返回每一个方向对应的数值
    """
    def __init__(self,mol:Mol) -> None:
        self.mol=mol
        self.chrg:Literal['mulliken','lowdin']='mulliken'
        self.prop:Literal['charge','spin']='charge'
        self.vects:list[np.ndarray]=None
        self.atoms:list[int]=None
        self.zero:bool=True
        self.keep:bool=False
        self.abs:bool=False

    def calculate(self):
        assert self.chrg is not None,'没有指定方法'
        assert self.prop is not None,'没有指定属性'
        assert self.vects is not None,'没有指定方向'
        assert self.atoms is not None,'没有指定原子'
        assert len(self.atoms)==len(self.vects),'方向与原子数量应该一致'
        # 获取投影后的系数轨道
        CM=self.mol.projCM(self.atoms,self.mol.O_obts,self.vects,self.zero,self.keep)
        CM_c=np.copy(self.mol.CM) # 先把原来的系数矩阵保留下来
        self.mol.props['CM']=CM # 替换为投影后的系数矩阵
        # 计算各种性质
        if self.prop=='charge':
            values=self.atomCharge(self.chrg)
        if self.prop=='spin':
            values=self.atomSpin(self.chrg)
        # 替换为原本的系数矩阵
        self.mol.props['CM']=CM_c 
        return values
    
    def atomCharge(self,chrg:str)->np.ndarray:
        """计算mulliken电子数"""
        caler=charge.Calculator(self.mol)
        caler.chrg=chrg
        charges=caler.calculate()
        atomics=np.array(self.mol.atoms.atomics)
        eleNums=atomics-charges
        idxs=[a-1 for a in self.atoms]
        return eleNums[idxs]

    def atomSpin(self,chrg:str)->np.ndarray:
        """计算mulliken自旋"""
        caler=spin.Calculator(self.mol)
        caler.chrg=chrg
        spins=caler.calculate()
        idxs=[a-1 for a in self.atoms]
        return spins[idxs]
    
    def atomEngs(self):
        """原子能量"""
        caler=energy.Calculator(self.mol)
        engs=caler.calculate()
        idxs=[a-1 for a in self.atoms]
        return engs[idxs]

    def freeValence(self):
        """计算原子自由价"""
        caler=freeValence.Calculator(self.mol)
        caler.calculate()

    def projVector(self)->np.ndarray:
        """轨道系数向量投影"""
        obts=self.mol.O_obts
        proj=np.zeros(shape=(len(self.atoms),len(obts),3)) #[原子/方向，轨道]
        for a,atm in enumerate(self.atoms):
            vect=self.vects[a]
            atom=self.mol.atom(atm)
            if atom.symbol=='H':
                proj[a,:]=0
                continue
            for o,obt in enumerate(obts):
                Cops=atom.get_pProj(vect,obt,False) # p轨道向量
                Cop=np.array(Cops).mean(axis=0)
                proj[a,o]=Cop
        return proj.sum(axis=1)
    
    def printRes(self):
        resStr=self.resStr()
        printer.info('方向电子自旋分布: ')
        printer.res(resStr)
        
    def resStr(self):
        values=self.calculate()
        return lutils.atomValueStr(self.mol,self.atoms,values)



from collections.abc import Iterator
from pywfn.base.atom import Atom
from pywfn import base
import numpy as np
# 如何获取一个键？
# Mol.getBond(a1,a2)
# Mol应该有bonds属性，是一个字典，索引是'a1-a2'和'a2-a1'都指向该键
from functools import cached_property
class Bond:
    def __init__(self,mol:"base.Mol",atm1:int,atm2:int) -> None:
        self.mol=mol
        self.atm1=atm1
        self.atm2=atm2
        self._length:float=None
        self.key:str=f'{atm1}-{atm2}'
        self.ats:list[int]=[atm1,atm2]
    
    @property
    def a1(self):
        return self.mol.atom(self.atm1)
    
    @property
    def a2(self):
        return self.mol.atom(self.atm2)

    @cached_property
    def length(self):
        """获取键长"""
        return np.linalg.norm(self.a2.coord-self.a1.coord)

    @cached_property
    def vector(self):
        """获取键向"""
        return self.a2.coord-self.a1.coord

    @cached_property
    def center(self):
        """获取键中心坐标"""
        return (self.a1.coord+self.a2.coord)/2
    
    @property
    def btype(self):
        """获取键类型
        1:单键
        2:双键
        3:三键
        4:芳香
        """
        blen=self.length
        if blen<1.38*1.889:
            return 1
        else:
            return 2
    
    def __repr__(self) -> str:
        return f'{self.a1.idx}-{self.a2.idx},{self.length:.4f}'
    
class Bonds:
    def __init__(self,mol:"base.Mol") -> None:
        self.mol=mol
        self.bonds:list[Bond]=[]
    
    def add(self,idx1:int,idx2:int):
        """添加一个键"""
        if idx1>=idx2:idx1,idx2=idx2,idx1
        bond=Bond(self.mol,idx1,idx2)
        self.bonds.append(bond)
        return bond
    
    def get(self,idx1:int,idx2:int):
        """获取一个键"""
        if idx1>=idx2:idx1,idx2=idx2,idx1
        for bond in self.bonds:
            if bond.a1.idx==idx1 and bond.a2.idx==idx2:
                return bond
        raise f'没有指定的键{idx1}-{idx2}'
    
    
    @property
    def keys(self):
        return [bond.key for bond in self.bonds]
    
    @property
    def num(self):
        return len(self.bonds)

    def __iter__(self) -> Iterator[Bond]:
        for bond in self.bonds:yield bond

    def __bool__(self)->bool:
        return len(self.bonds)!=0


"""
需要导出文件：gjf等
"""
from pathlib import Path
from pywfn.data import temps

def gjf(path,chk:str,coord:list[list[float]],charge:int,spin:int):
    temp=temps.gjf
    replaces=[
        ['<COORD>', cord2str(coord)],
        ['<CHK>', chk],
        ['<CHARGE>',f'{charge}'],
        ['<MULTI>',f'{spin}']
    ]
    for k,v in replaces:
        temp=temp.replace(k,v)
    Path(path).write_text(temp)

def cord2str(coord:list[list[str,float,float,float]]):
    strs=[]
    for s,x,y,z in coord:
        strs.append(f' {s:<6}{x:>14.8f}{y:>14.8f}{z:>14.8f}')
    return '\n'.join(strs)


"""
该脚本计算指定原子的pi电子数
"""
from pywfn.base import Mol,Atom
from pywfn.maths import CM2PM
from pywfn.utils import printer
from pywfn.atomprop import lutils,dirProps,AtomCaler
from typing import Literal


class Calculator(AtomCaler):
    def __init__(self,mol:Mol) -> None:
        self.mol=mol
        self.caler=dirProps.Calculator(mol)
        self.chrg:Literal['mulliken','lowdin']='mulliken'
        self.prop:Literal['charge','spin']='charge'
        self.atoms:list[int]=None
        self.logTip:str=''

    def calculate(self):
        """
        计算所有原子的π电子
        如果指定方向，则计算方向电子
        每个原子都有应该根据自己的法向量方向来计算(当法向量都相同时相当于都一样)
        """
        atoms=self.mol.heavyAtoms
        atoms=lutils.atomIdxs(atoms)
        vects=lutils.get_vects(self.mol,atoms) # 每个原子都获得自己的法向量

        self.caler.chrg=self.chrg
        self.caler.prop=self.prop
        self.caler.vects=vects
        self.caler.atoms=atoms
        values=self.caler.calculate()
        return values

    def print(self,resStr:str):
        printer.info('所有非H原子的pi电子分布: ')
        printer.res(resStr)
    
    def resStr(self):
        elects=self.calculate()
        atoms=lutils.atomIdxs(self.mol.atoms)
        resStr=lutils.atomValueStr(self.mol,atoms,elects)
        resStr+=f'\ntotal: {sum(elects):.4f}'
        return resStr


from collections.abc import Iterator
import numpy as np
from functools import cached_property,lru_cache

from pywfn import config
from pywfn import base
from pywfn import maths

from pywfn.utils import printer

"""
一个原子的轨道组合系数就是一个矩阵，行数是基函数的数量，列数是分子轨道的数量
与轨道系数相关的太多了,尽量精简
OC是该原子的轨道系数
OCI为某一列的原子轨道系数,P:bool参数用来确定是否只要P轨道的系数
一个人原子的属性应该是标准的
原子轨道layer:List[str]
原子轨道系数OC:np.ndarray
读取器直接设置这些属性
"""
from pywfn.data.elements import elements
class Atom:
    def __init__(self,symbol:str,coord:np.ndarray,idx:int,mol:"base.Mol"): # 每个原子应该知道自己属于哪个分子
        self.symbol=symbol
        self.coord=coord
        self.coord.flags.writeable=False
        self.idx=idx
        self.mol:"base.Mol"=mol

        self._layersData={}
        self.layers:list[str]=None
        self._squareSum=None
        self._sContribution:dict={}
        self.OC:np.ndarray=None
    

    @cached_property
    def atomic(self)->int:
        return elements[self.symbol].idx

    @cached_property
    def obtBorder(self): # 获取元素上下界
        idxs=[i for i,idx in enumerate(self.mol.obtAtms) if idx==self.idx]
        return idxs[0],idxs[-1]+1 #因为最后一个元素不包含
    
    @property
    def obtCoeffs(self):
        """获取该原子对应的系数"""
        u,l=self.obtBorder
        return self.mol.CM[u:l,:]

    @cached_property
    def neighbors(self)->list[int]:
        """每个原子相邻的原子有哪些，根据分子的键来判断"""
        idxs=[]
        bonds=self.mol.bonds
        for bond in bonds:
            idx1,idx2=bond.ats
            if idx1==self.idx:idxs.append(idx2) # 添加不是该原子的键上的原子
            if idx2==self.idx:idxs.append(idx1)
        return [idx for idx in set(idxs)]
    
    def get_wfnv(self,coords:np.ndarray,obt:int,range_:float=config.RENDER_ATOM_RANGE)->np.ndarray:
        """
        计算波函数值
        pos：原子中心为原点的笛卡尔坐标[n,3]
        obt：分子轨道序数
        range_：计算的范围
        """
        dis=np.linalg.norm(coords,axis=1) #所有点距离原子的距离
        idx=np.argwhere(dis<range_).flatten() # 挑选出距离小于指定距离的点,根据距离少渲染一些点可以加快计算速度
        wfnv=np.zeros(len(coords))
        wfni=self.mol.gto.agto(coords[idx,:],self.idx,obt)
        wfnv[idx]=wfni
        return wfnv

    def get_dens(self,obts:list[int],coords:np.ndarray):
        """
        计算电子密度
        obts:要计算的分子轨道
        coords:空间笛卡尔坐标[n,3]，原子中心为原点
        """
        atmDens=np.zeros(shape=(len(coords)))
        for obt in obts:
            oE=self.mol.oE # 轨道占据数0/1/2
            wfnv=self.get_wfnv(coords,obt)
            dens=oE*wfnv**2 # 波函数的平方乘轨道占据数
            atmDens+=dens
        return atmDens

    @lru_cache
    def pLayersCs(self,obt:int):
        '''获取原子某一轨道的p层数据'''
        a,b=self.obtBorder
        layers=self.mol.obtSyms[a:b]
        pIndex=[i for i,l in enumerate(layers) if 'P' in l]
        return self.obtCoeffs[pIndex,obt]
    
    def get_pProj(self,direct:np.ndarray,obt:int)->list[np.ndarray]:
        """计算原子p系数在某个方向上的投影,返回n个三维向量"""
        assert isinstance(direct,np.ndarray),"方向需要为np.ndarray"
        assert direct.shape==(3,),'方向向量长度应该为3'
        length=np.linalg.norm(direct)
        assert np.abs(length-1)<1e-4,'方向范数应该为1'
        assert length!=0,"方向向量长度不能为0"
        cs=self.pLayersCs(obt) # p轨道的系数
        assert len(cs)%3==0,"p轨道系数的长度应为3n"
        ps=[np.array(cs[i:i+3]) for i in range(0,len(cs),3)] #每一项都是长度为3的数组
        lens=np.dot(ps,direct)
        ps_=[l*direct for l in lens] # 轨道向量在法向量方向上的投影
        return ps_
    
    def get_sProj(self,direct:np.ndarray,obt:int):
        """计算原子p系数在某个方向上的投影,返回n个三维向量"""
        assert isinstance(direct,np.ndarray),"方向需要为np.ndarray"
        assert np.linalg.norm(direct)!=0,"方向向量长度不能为0"
        direct/=np.linalg.norm(direct) # 投影向量归一化
        a,b=self.obtBorder
        layers=self.mol.obtAngs[a:b]
        sIndex=[i for i,l in enumerate(layers) if 'S' in l]
        return self.obtCoeffs[sIndex,obt]
    
    
    @lru_cache
    def get_Normal(self,aroundIdx:int=None,main:bool=True)->None|np.ndarray: # 一个原子应该知道自己的法向量是多少
        """
        mian:代表是否为主动调用(防止递归)
        获取原子的法向量,垂直于键轴,如果不传入另一个原子,则垂直于三个原子确定的平面\n
        1.如果该原子连接两个原子
            1.1 如果两个原子不在一条直线上,有法向量，直接返回\n
            2.2 如果两个原子在一条之下上则没有法向量
        2.如果该原子没有法向量\n
            2.1 如果相邻的原子有法向量,返回邻原子的法向量\n
            2.2 如果相邻原子没有法向量,返回None
        """

        locNum=0.001
        neighbors=self.neighbors
        normal=None
        if len(neighbors)==3:
            if aroundIdx is None: #如果传入的没有相邻原子的序号
                #获取中心原子到三个相邻原子的法向量
                p1,p2,p3=[(each.coord-self.coord) for each in neighbors] 
            else:
                p1,p2,p3=[(each.coord-self.coord)*(1 if each.idx==aroundIdx else locNum) for each in neighbors] 
            normal=maths.get_normalVector(p1,p2,p3)
        elif len(neighbors)==2:
            p2=self.coord
            p1,p3=neighbors[0].coord,neighbors[1].coord
            angle=maths.vector_angle(p1-p2,p3-p2)
            if 0.2<=angle<=0.8:
                normal=maths.get_normalVector(p1,p2,p3,linear=True)
            else:
                pass
        elif len(neighbors)==1:
            a2=neighbors[0]
            normal = a2.get_Normal(self.idx)
        else:
            pass
        # elif main: #如果是在递归中调用本函数的话就不要再次递归了
        #     for each in neighbors:
        #         normal_=each.get_Normal(self.idx,main=False)
        #         if normal_ is not None:
        #             normal=normal_
        #             return normal
        #         return None #如果周围原子也没有法向量的话，返回None
        if normal is None:
            printer.warn(f'原子: {self.idx} 没有法向量')
        else:
            if maths.vector_angle(normal,config.BASE_VECTOR)>0.5:
                normal*=-1
        return normal


    @lru_cache
    def get_vertObt(self,idx1:int,idx2:int)->np.ndarray:
        """从HOMO轨道开始寻找垂直于键轴的轨道方向"""
        obts=self.mol.O_obts
        bondVector=self.mol.atom(idx1).coord-self.mol.atom(idx2).coord
        for obt in obts[::-1]:
            obtWay=self.get_obtWay(obt)
            if obtWay is None:continue
            printer.info(f'以{obt}号轨道方向作为法向量')
            if maths.vector_angle(bondVector,obtWay,trans=True)>0.4:
                if maths.vector_angle(obtWay,config.BASE_VECTOR)>0.5:obtWay*=-1
                return obtWay
    
    def get_projWay(self):
        """
        获得系数投影方向
        有法向量用法向量,没有法向量用垂直于轨道和键轴的方向
        没有法向量的时候,相当于原子在线性区域,两个键平行,但还是用平均值比较合理
        """
        norm=self.get_Normal()
        if norm is not None:
            return norm
        else:
            a1,a2=self.neighbors #相邻两个原子
            return self.get_vertObt(a1,a2) #返回垂直于键轴的轨道方向
    
    @lru_cache
    def get_obtWay(self,obt:int)->np.ndarray:
        """获得原子某一轨道的方向"""
        maxPos,maxValue=maths.get_extraValue(self,obt)
        way=maxPos # 如果两者相同说明完全没有电子云、这显然不对啊
        if np.linalg.norm(way)==0:
            printer.warn(f'轨道{obt}无方向')
            return None
        return way/np.linalg.norm(way)

    def __repr__(self) -> str:
        x,y,z=self.coord
        return f'{self.idx} {self.symbol} ({x:.4f},{y:.4f},{z:.4f})'

    @lru_cache
    def get_sCont(self,orbital:int):
        """获取某个原子轨道的贡献"""
        s=self.OC[0,orbital]
        contribution=s**2/self.mol.As[orbital]
        return contribution

    def get_direct(self):
        """
        获取反应方向
        对于两个键的，为平面内夹角相同方向
        
        """
        


class Atoms:
    def __init__(self,mol:"base.Mol") -> None:
        self.mol=mol
        self.atoms:list[Atom]=[]
    
    def __repr__(self) -> str:
        atoms=[atom.symbol for atom in self.atoms]
        return f'Atoms:{len(self.atoms)}\n'+' '.join(atoms)
    
    def add(self,symbol:str,coord:np.ndarray):
        atom=Atom(symbol,coord,len(self.atoms)+1,self.mol)
        self.atoms.append(atom)
    
    @property
    def atomics(self)->list[int]:
        return [a.atomic for a in self.atoms]
    
    @property
    def symbols(self)->list[str]:
        return [a.symbol for a in self.atoms]
    
    
    @property
    def indexs(self)->list[int]:
        return [a.idx for a in self.atoms]
    
    @property
    def num(self)->int:
        return len(self.atoms)

    def __getitem__(self,item)->Atom:
        return self.atoms[item]
    
    def __iter__(self)->Iterator[Atom]:
        for atom in self.atoms:
            yield atom
        
    def __bool__(self)->bool:
        return len(self.atoms)!=0
    
    def __len__(self)->int:
        return len(self.atoms)


"""
三个分子的差值确定的属性
- 福井函数
- parr函数
"""
from pywfn.base import Mol
from pywfn.atomprop import charge,dirProps,AtomCaler,energy, spin
import numpy as np
from typing import Literal

class Calculator(AtomCaler):
    def __init__(self,mols:list[Mol]) -> None:
        self.mols=mols
        self.natm=len(mols[0].atoms)
        self.vects=None # 可以计算方向属性
        self.atoms=None
        self.chrg=None # 电荷类型
        self.prop=None # 计算属性
        self.func:Literal['fukui','parr']=None # 函数类型
        self.calers=[]
        self.hists={}
        
    
    def calculate(self)->np.ndarray:
        funcs={'fukui':self.fukuiFunc,'parr':self.parrFunc}
        if self.vects:
            assert self.vects is not None,'没有指定方向'
            self.calers=[dirProps.Calculator(mol) for mol in self.mols]
            for caler in self.calers:
                caler.atoms=self.atoms
                caler.vects=self.vects
                caler.chrg=self.chrg
                caler.prop=self.prop
            return funcs[self.func]
        else:
            if self.func=='fukui':
                self.calers=[charge.Calculator(mol) for mol in self.mols]
            if self.func=='parr':
                self.calers=[spin.Calculator(mol) for mol in self.mols]
            return funcs[self.func]

    def setProject(self,stype:bool):
        """
        检查是否需要投影
        stype,修改(T)或撤回修改(F)
        """
        if self.vects is None:return
        for m,mol in enumerate(self.mols):
            if stype:
                CMo=mol.CM
                self.hists[f'CM_{m}']=CMo # 将老的存储下来
                CMn=mol.projCM(self.atoms,mol.O_obts,self.vects,True,False)
                mol.datas['CM']=CMn
            else:
                CMo=self.hists[f'CM_{m}'] # 恢复成老的
                mol.datas['CM']=CMo
    
    def fukuiFunc(self)->np.ndarray:
        self.setProject(True)
        assert len(self.mols)==3,'需要三个分子'
        charges=np.zeros(shape=(self.natm,3))
        dcharge=np.zeros(shape=(self.natm,2))
        for c,caler in enumerate(self.calers):
            charge=caler.calculate()
            charges[:,c]=charge
        dcharge[:,0]=-(charges[:,0]-charges[:,1])
        dcharge[:,1]=-(charges[:,1]-charges[:,2])
        self.setProject(False)
        return dcharge
    
    def parrFunc(self)->np.ndarray:
        self.setProject(True)
        assert len(self.mols)==2,'需要两个分子'
        spins=np.zeros(shape=(self.natm,2))
        for c,caler in enumerate(self.calers):
            spin=caler.calculate()
            spins[:,c]=spin
        self.setProject(False)
        return spins
    
    def atomEngs(self)->np.ndarray:
        """计算原子轨道能"""
        self.setProject(True)
        self.calers=[energy.Calculator(mol) for mol in self.mols]
        atmNum=len(self.mols[0].atoms)
        allEngs=np.zeros(shape=(atmNum,3))
        for c,caler in enumerate(self.calers):
            engs=caler.calculate()
            allEngs[:,c]=engs
        self.allEngs=allEngs
        delEngs=np.zeros(shape=(atmNum,2))
        delEngs[:,0]=allEngs[:,0]-allEngs[:,1]
        delEngs[:,1]=allEngs[:,1]-allEngs[:,2]
        self.setProject(False)
        return delEngs
    


"""
计算空间波函数
"""
from pywfn.base import Mol
from pywfn.maths import Gto
import numpy as np

from pywfn.data import sphGrid
weight = sphGrid.gridData[:, -1]
coords = sphGrid.gridData[:, :3]

class Calculator:
    def __init__(self,mol) -> None:
        self.mol:Mol=mol
        self.pos=np.zeros((1,3)) # 初始坐标设为原点

    def molDens(self,pos:np.ndarray):
        """计算分子电子密度"""
        obts = self.mol.O_obts
        dens=np.zeros(len(pos))
        for o in obts:
            wfn=self.obtWfn(o,pos)
            dens+=wfn**2*self.mol.oE
        return dens
    
    def molDens2(self,pos:np.ndarray):
        dens=np.zeros(len(pos))
        for atom in self.mol.atoms:
            dens+=self.atmDens(atom.idx,pos)
        return dens
    
    def atmDens(self,atm:int,pos:np.ndarray):
        """计算原子电子密度"""
        nmat=self.mol.CM.shape[0]
        atom=self.mol.atom(atm)
        dens=np.zeros(len(pos))
        u,l = atom.obtBorder
        for i in range(nmat):
            wfn_i=self.atoWfn(i,pos)
            for j in range(u,l):
                wfn_j=self.atoWfn(j,pos)
                dens+=wfn_i*wfn_j*self.mol.PM[i,j]
        return dens
    
    def obtWfn(self,obt:int,pos:np.ndarray):
        """
        计算分子轨道的波函数，为原子轨道的线性组合
        obt：分子轨道指标
        coefs：线性组合系数
        """
        wfn=np.zeros(len(pos))
        coefs=self.mol.CM[:,obt]
        for c,coef in enumerate(coefs):
            wfn+=coef*self.atoWfn(c,pos)
        return wfn
    
    def atoWfn(self,i:int,pos:np.ndarray):
        """
        计算原子轨道的波函数，形成在轨道线性组合之前
        i:原子轨道指标
        """
        atms = self.mol.obtAtms
        shls = self.mol.obtShls
        syms = self.mol.obtSyms
        lmns = [self.mol.basis.sym2lmn(sym) for sym in syms]
        
        lmn = lmns[i]
        atm = atms[i]
        shl = shls[i]
        ang = sum(lmn)
        atmic = self.mol.atom(atm).atomic
        basis = self.mol.basis.get(atmic, shl, ang)
        exps = [b.exp for b in basis]
        coes = [b.coe for b in basis]
        pos_ = pos-self.mol.atom(atm).coord # 空间坐标-原子坐标=以原子为中心的空间坐标
        R2 = np.sum(pos_**2, axis=1)
        wfn = self.mol.gto.cgf(exps, coes, lmn, R2, pos_)  # 空间坐标-以原子为中心的坐标
        return wfn



"""
键级也不止一种，都在这里实现吧
"""
from pywfn.base import Mol
from pywfn.atomprop import direction
from pywfn.maths import CM2PM
from pywfn.utils import printer
import numpy as np
from itertools import product

class Calculator:
    def __init__(self,mol:Mol) -> None:
        self.mol=mol

    def mayer(self,*,PM=None,bonds=None)->np.ndarray:
        """计算指定键的mayer键级，可以有多个"""
        # 获取密度矩阵 P
        if PM is None:
            PM=self.mol.PM
        # 获取重叠矩阵
        SM=self.mol.SM
        PS=PM@SM
        OM=PS*PS.T
        orders=[]
        if bonds is None:
            bonds=[bond.ats for bond in self.mol.bonds]
        for a1,a2 in bonds:
            u1,l1=self.mol.atom(a1).obtBorder
            u2,l2=self.mol.atom(a2).obtBorder
            order=np.sum(OM[u1:l1,u2:l2])
            orders.append([a1,a2,order])
        order = np.array(orders)
        # print(order)
        return order
    
    def dirMayer(self,bonds:list[list[int,int]])->np.ndarray:
        """带有方向的Mayer键级[d,6](a1,a2,x,y,z,v)"""
        dirCaler=direction.Calculator(self.mol)
        obts=self.mol.O_obts
        result=[]
        for a1,a2 in bonds:
            dirs=dirCaler.reaction(a1)
            atms=[a1]*len(dirs)
            if a1>a2:a1,a2=a2,a1
            for d,(atm,dir_) in enumerate(zip(atms,dirs)):
                CMp=self.mol.projCM(obts,[atm],[dir_],True,True)
                PMp=CM2PM(CMp,self.mol.O_obts,self.mol.oE)
                a1_,a2_,order=self.mayer(PM=PMp,bonds=[[a1,a2]]).flatten()
                assert a1==a1_ and a2==a2_,"原子不对应"
                x,y,z=dir_
                result.append([a1,a2,x,y,z,order])
        return np.array(result)
    
    def piOrder(self):
        """
        计算pi键级，每一个可能的π键计算出一个π键级
        """
        dirCaler=direction.Calculator(self.mol)
        atms=[]
        dirs=[]
        for atom in self.mol.atoms:
            normal=dirCaler.normal(atom.idx) # 原子的法向量
            if normal is None:continue
            atms.append(atom.idx)
            dirs.append(normal)
        PMp=self.mol.projCM(self.mol.O_obts,atms,dirs,False,False)
        PMp=CM2PM(PMp,self.mol.O_obts,self.mol.oE)
        result=self.mayer(PM=PMp)
        values=np.sqrt(result[:,-1])
        indexs=values>0
        result[:,-1]=values
        return result[indexs,:]

    def hmo(self):
        # 1.建立键连矩阵
        atms=self.mol.heavyAtoms
        natm=len(atms)
        BM=np.zeros(shape=(natm,natm)) # 键连矩阵
        DM=np.zeros_like(BM) # 键长矩阵
        for i,j in product(range(natm),range(natm)):
            a1,a2=atms[i],atms[j]
            if a1>=a2:continue
            bond=self.mol.atom(a1).coord-self.mol.atom(a2).coord
            dist=np.linalg.norm(bond)
            DM[i,j]=dist
            DM[j,i]=dist
            if dist>1.7*1.889:continue
            BM[i,j]=1.0
            BM[j,i]=1.0
        # 2.求解
        e,C=np.linalg.eigh(BM) # 矩阵对角化
        nele=int(len(atms)-self.mol.charge) #电子数量
        idxs=np.argsort(e)[:nele//2] # 占据轨道
        CM=C[:,idxs] # 每一列对应一个特征向量
        # 3.构建键级矩阵
        result=[]
        OM=np.zeros_like(BM)
        for i,j in product(range(natm),range(natm)):
            order=np.sum(CM[i,:]*CM[j,:])*2
            OM[i,j]=order
            if i>=j:continue
            if DM[i,j]>1.7*1.889:continue
            result.append([atms[i],atms[j],order])
        result=np.array(result)
        # vals=result[:,-1]
        # vals=np.sqrt(vals**2)
        # result[:,-1]=vals
        return np.abs(result)

    def multiCenter(self,atms:list[int]):
        """计算多中心键级"""
        pass


"""
基于python的波函数分析程序
"""
import numpy as np
np.set_printoptions(precision=4, suppress=True)

# __version__='0.0.1'
# __description__='A wave function analysis program based on python'
# __author__='XiaoFei Shi'
# __email__='1103275712@qq.com'
# __url__='https://www.xiaofei911.top/mkdocs/pywfn/'

def main():
    from pywfn.shell import Shell
    Shell().homePage()


"""
mol文件读取器
"""
import numpy as np
from numpy import ndarray
from pywfn.reader import Reader
from functools import lru_cache
class MolReader(Reader):
    def __init__(self, path: str) -> None:
        super().__init__(path)
        line3=self.getline(3)
        self.natm=int(line3[:3])
        print(self.natm)

    def get_coords(self) -> ndarray:
        return self.read_coords()[1]
    
    def get_symbols(self) -> list[str]:
        return self.read_coords()[0]
    
    @lru_cache
    def read_coords(self):
        """读取原子坐标"""
        xyzs=[]
        syms=[]
        for i in range(4,4+self.natm):
            line=self.getline(i)[:-1]
            x=line[ 0:10].strip()
            y=line[10:20].strip()
            z=line[20:30].strip()
            s=line[30:32].strip()
            xyzs.append([x,y,z])
            syms.append(s)
        return syms,np.array(xyzs,dtype=float)*1.889



"""
计算与分子轨道对应的电子分布
分子轨道中每个基函数的电子数量
"""
from pywfn.base import Mol
import numpy as np
from pywfn.utils import printer

class Calculator:
    def __init__(self,mol:Mol) -> None:
        self.mol=mol
        self.CM=mol.CM
        self.NM:np.ndarray=None
    
    def calculate(self)->np.ndarray:
        """
        返回与轨道系数矩阵对应的电子分布矩阵
        """
        CM=self.CM
        SM=self.mol.SM
        row,col=CM.shape
        oE=self.mol.oE
        NM=np.zeros(shape=(row,row,col))
        for a in range(row):
            for i in range(row):
                for j in self.mol.O_obts:
                    # 起始系数为0的可以直接跳过
                    NM[a,i,j]=CM[a,j]*CM[i,j]*SM[i,a]*oE
        # NM=NM.sum(axis=1)
        NM:np.ndarray=np.sum(NM,axis=1)[:,self.mol.O_obts] # 二维矩阵[n,occ]
        self.NM=NM
        NMo:np.ndarray=NM.sum(axis=0) # 每个轨道的电子数

        obtEngs=np.array(self.mol.obtEngs).reshape(1,-1)[:,self.mol.O_obts]
        # print(obtEngs)
        EM=NM*obtEngs
        return EM


"""
定义gjf的读取器，感觉可能也不会用得到
"""
from numpy import ndarray
from pywfn.reader import lutils
import re
import numpy as np
from functools import lru_cache

class GjfReader(lutils.Reader):
    def __init__(self,path) -> None:
        super().__init__(path)
    
    def get_symbols(self) -> list[str]:
        symbols,coords=self.read_coord()
        return symbols
    
    def get_coords(self) -> ndarray:
        symbols,coords=self.read_coord()
        return coords
    
    def get_charge(self) -> int:
        charge,multi=self.read_multi()
        return charge
    
    def get_spin(self) -> int:
        charge,multi=self.read_multi()
        return multi

    @lru_cache
    def read_coord(self):
        finds=re.findall(' ([A-Za-z]+) +(-?\d+.\d+) +(-?\d+.\d+) +(-?\d+.\d+)',self.text)
        symbols=[]
        coords=[]
        for s,x,y,z in finds:
            symbols.append(s)
            coords.append([x,y,z])
        return symbols,np.array(coords,dtype=float)
    
    @lru_cache
    def read_multi(self)->tuple[int]:
        """读取电荷与自旋多重度"""
        finds=re.search('(-?\d) (\d)',self.text)
        assert finds is not None,'未正确读取电荷与自旋多重度!'
        charge,multi=finds.groups()
        return int(charge),int(multi)


"""
轨道挑选法+HMO键级公式
"""
import numpy as np

from pywfn.base import Mol
from pywfn.bondprop import lutils,Caler

class Calculator(Caler):
    def __init__(self,mol:Mol) -> None:
        self.mol=mol
        self.oe=self.mol.oE
        self.bond:list[int]=None

    def calculate(self):
        idx1,idx2=self.bond
        cenAtom=self.mol.atom(idx1)
        aroAtom=self.mol.atom(idx2)
        obts=self.mol.O_obts

        As=self.mol.As[obts]
        piUnits=[lutils.judgeOrbital(cenAtom,aroAtom,obt,cenAtom.get_Normal()) for obt in obts]
        cenRes=np.sqrt(np.sum(cenAtom.OC[:,obts]**2,axis=0))/As
        aroRes=np.sqrt(np.sum(aroAtom.OC[:,obts]**2,axis=0))/As
    
        orders=cenRes*aroRes*np.array(piUnits)*self.mol.oE
        lutils.printOrders(orders,[obt+1 for obt in obts])
        return sum(orders)


"""
定义所有读取类的基类
默认读取的所有信息返回为空
不同类型的文件覆盖不同的读取函数

可以为每个分子对象分配一个Reader对象
用户可以自己分配分子属性，也可以从文件中读取
当调用分子属性且未分配时，会通过reader对象读取属性，若读取失败则报错
每一个方法都需要被子例覆盖

- 原子类型
- 原子坐标
- 轨道系数，[n,n]/[n,2n]
- 分子能量
- 原子轨道所属原子
- 原子轨道角动量类型
- 分子轨道能量
- 分子轨道占据类型 占据|非占据
- 重叠矩阵
- 波函数类型 开壳层/闭壳层
"""
from pywfn import base
from pywfn import data
import numpy as np
from pathlib import Path
from functools import cached_property
import linecache
from pywfn.utils import printer

class Reader:
    def __init__(self,path:str) -> None:
        self.path:str=path
    
    @cached_property
    def text(self)->str:
        return Path(self.path).read_text()

    @property
    def fileName(self)->str:
        return Path(self.path).name
    
    @cached_property
    def lineNum(self)->int:
        """获取文件行数"""
        lineNum=0
        with open(self.path,'r') as file:
            for _ in file:
                lineNum+=1
        return lineNum
    
    def getline(self,idx:int,keepEnd:bool=True)->str:
        if keepEnd:
            return linecache.getline(self.path,idx+1)
        else:
            return linecache.getline(self.path,idx+1).replace('\n','')
    
    def getlines(self,idx1:int,idx2:int)->list[str]:
        lines=[]
        for i in range(idx1,idx2):
            lines.append(self.getline(i))
        return lines


    def get_coords(self)->np.ndarray:
        """原子坐标[n,3]"""
        raise

    def get_symbols(self)->list[str]:
        """原子符号[n]"""
        raise

    def get_energy(self)->float:
        """获取分子能量"""
        raise

    def get_charge(self)->int:
        """获取分子电荷"""
        printer.warn("无法读取电荷")
        return None

    def get_spin(self)->int:
        """获取分子自旋"""
        printer.warn("无法读取自旋")
        return None

    def get_CM(self)->np.ndarray:
        """获取系数矩阵[m,m]/[m,2m]"""
        raise

    def get_SM(self)->np.ndarray:
        """获取重叠矩阵[m,m]"""
        raise

    def get_obtEngs(self)->list[float]:
        """获取分子轨道能量[m]"""
        raise

    def get_obtOccs(self)->list[bool]:
        """获取轨道类型，占据|非占据[m]"""
        raise

    def get_obtAtms(self)->list[int]:
        """获取轨道系数每一行对应的原子[m]"""
        raise

    def get_obtShls(self)->list[int]:
        """获取轨道系数每一行对应的原子层[m]"""
        raise
    
    def get_obtSyms(self)->list[float]:
        """获取轨道系数每一行对应的轨道符号(S,PX,PY...)"""
    
    def get_basis(self)->"data.Basis":
        """获取基组数据[m,4]
        元素,层数,角动量,指数,系数
        """
        raise


"""
公用的函数可以移到此处，以较少代码量
"""
from pywfn.base import Atom,Mol
import numpy as np
from pywfn import maths


def atomIdxs(atoms:list[Atom])->list[int]:
    return [atom.idx for atom in atoms]

def get_vects(mol:Mol,atoms:list[int]=None):
    """获取指定原子的法向量或与相邻的原子的法向量"""
    atoms:list[Atom]=[mol.atom(a) for a in atoms]
    vects=[]
    for atom in atoms:
        vect=atom.get_Normal()
        if vect is None:
            idxn=maths.search_sp2(atom.idx,mol)
            vect=mol.atom(idxn).get_Normal()
        vects.append(vect)
    return vects

def get_ects(mol:Mol,obts:list[int],CM:np.ndarray)->list[int]:
    """
    计算电子数量,如果不指定则计算所有的π电子
    如果指定方向，则计算方向电子
    每个原子都有应该根据自己的法向量方向来计算(当法向量都相同时相当于都一样)
    """
    # 计算密度矩阵
    PM=maths.CM2PM(CM,obts,mol.oE) # 计算密度矩阵
    # PS=PM*mol.SM
    # PSS=PS.sum(axis=0)
    # 矩阵乘法的迹的加和=矩阵对应元素乘积之和
    PS=PM@mol.SM
    EV=np.diagonal(PS)
    elects=[]
    for atom in mol.atoms:
        a1,a2=atom.obtBorder
        elect=EV[a1:a2].sum()
        elects.append(elect)
    return elects

def atomValueStr(mol:Mol,satoms:list[int],values:list[float]):
    resStr=''
    if values is None:return '非开窍层无法计算自旋!!'
    
    for i,atom in enumerate(mol.atoms):
        if atom.idx in satoms:
            idx=satoms.index(atom.idx)
            value=values[idx]
        else:
            continue
        resStr+=f'{atom.idx:<2}{atom.symbol:>2}{value:>15.8f}\n'

    return resStr


"""
计算与键相关的方向
"""
from pywfn.base import Mol
from pywfn.maths import points_rotate
import numpy as np

class Calculator:
    def __init__(self,mol:Mol) -> None:
        self.mol=mol

    def verticals(self,atm1,atm2)->np.ndarray:
        """计算垂直于键轴的方向"""
        bond=self.mol.bonds.get(atm1,atm2)
        bdir=bond.vector # 键轴向量
        anyv=np.random.rand(3) # 随机生成向量

        bdir/=np.linalg.norm(bdir)
        anyv/=np.linalg.norm(anyv)

        vect=np.cross(bdir,anyv) # 垂直于键轴的向量
        cent=self.mol.atom(atm1).coord
        point=cent+vect
        point=point.reshape(1,3)

        angs=np.linspace(0,np.pi*2,35,endpoint=False)
        result=np.zeros(shape=(len(angs),3))
        for a,ang in enumerate(angs):
            result[a]=points_rotate(point,cent,bdir,ang)-cent
        
        return result



from pywfn.base import Mol,Atom
import numpy as np
from pywfn.data import Elements
from functools import lru_cache
elements=Elements()
from pywfn.utils import printer
from pywfn.atomprop import lutils,AtomCaler
from pywfn import maths
from typing import Literal
from pywfn.maths import CM2PM

Chrgs=Literal['mulliken','lowdin','hirshfeld']

class Calculator(AtomCaler):
    def __init__(self,mol:"Mol"):
        self.logTip:str=''
        self.mol=mol
        self.chrg:Chrgs='mulliken'
    
    def calculate(self,chrg:Chrgs)->np.ndarray:
        if chrg=='mulliken':
            return self.mulliken()
        if chrg=='lowdin':
            return self.lowdin()
        if chrg=='hirshfeld':
            return self.hirshfeld()
    
    def mulliken(self,num:bool=False,PM:np.ndarray=None):
        """
        计算目录mulliken电荷
        num：是否只保留电子数
        """
        # 计算密度矩阵
        self.logTip='Mulliken电荷分布'
        if PM is None:PM=self.mol.PM
        # 矩阵乘法的迹的加和=矩阵对应元素乘积之和
        PS=PM@self.mol.SM
        EV=np.diagonal(PS)
        atoms=self.mol.atoms
        charges=np.zeros(len(atoms))
        for a,atom in enumerate(atoms):
            a1,a2=atom.obtBorder
            elect=EV[a1:a2].sum()
            if num:
                charges[a]=elect
            else:
                charges[a]=atom.atomic-elect
        return charges
    
    def lowdin(self,num:bool=False,PM:np.ndarray=None):
        """
        计算每个原子的lowdin电荷
        """
        self.logTip='Lowdin电荷分布'
        if PM is None:PM=self.mol.PM
        SM=self.mol.SM
        # 计算矩阵的1/2次方
        v, Q=np.linalg.eig(SM)
        V=np.diag(v)
        V_=V**0.5
        Q_=np.linalg.inv(Q)
        SM_half=Q@(V_@Q_)
        SPS=SM_half@(PM@SM_half)
        eleNums=np.diagonal(SPS)
        charges=np.zeros(len(self.mol.atoms))
        for a,atom in enumerate(self.mol.atoms):
            u,l=atom.obtBorder
            eleNum=eleNums[u:l].sum()
            if num:
                charges[a]=eleNum
            else:
                charges[a]=atom.atomic-eleNum
        return charges
    
    def hirshfeld(self):
        """
        计算原子的Hirshfeld电荷，目前还未成功（分子的电子密度计算不正确）
        """
        from pywfn.spaceProp import wfn
        wfnCaler=wfn.Calculator(self.mol)
        self.mol.bohr=True
        from pywfn.data import sphGrid,radDens
        coords=sphGrid.gridData[:,:3]# 原点为0的坐标
        weight=sphGrid.gridData[:,3]
        npos=len(coords)
        atoms:list[int]=self.mol.atoms.indexs
        chargs=np.zeros(shape=(len(atoms)))
        for a1,atom1 in enumerate(self.mol.atoms): # 计算每一个原子的电荷
            pos=coords+atom1.coord # 该原子周围点的空间坐标
            # molDens=self.mol.get_dens(atoms,obts,coord)*weight # 计算分子的电子密度
            molDens=np.zeros(npos) # 计算分子的电子密度
            proDens=np.zeros(npos)
            for a2,atom2 in enumerate(self.mol.atoms): #计算前置分子密度
                radius=np.linalg.norm(pos-atom2.coord,axis=1) #所有坐标对应的半径
                dens1=radDens.get_radDens(atom2.atomic,radius)*weight # 插值法计算提前存储好的密度
                proDens+=dens1
                if a2==a1:atmDens=dens1
                dens2=wfnCaler.atmDens(atom2.idx,pos-atom2.coord)*weight
                molDens+=dens2
                print(f'{dens1.sum():.4f},{dens2.sum():.4f}')
            ratio=np.divide(atmDens,proDens,out=np.zeros_like(atmDens),where=proDens!=0)
            chargs[a1]=np.sum(ratio*(molDens-proDens))
            atmQ,proQ,molQ=np.sum(atmDens),np.sum(proDens),np.sum(molDens)
            print(f'{atmQ=:.4f},{proQ=:.4f},{molQ=:.4f}')
        chargs=[atom.atomic-val for atom,val in zip(self.mol.atoms,chargs)]
        return chargs
    
    def hirshfeld2(self):
        """第二种计算hirshfeld电荷的方法"""
        from pywfn.data import sphGrid,radDens
        from pywfn.spaceProp import wfn
        coords=sphGrid.gridData[:,:3]# 原点为0的坐标
        weight=sphGrid.gridData[:,3]
        pos=[]
        wei=[]
        for atom in self.mol.atoms:
            pos.append(coords+atom.coord)
            wei.append(weight)
        pos=np.vstack(pos)
        natm=len(self.mol.atoms)
        wei=np.hstack(wei)/natm
        
        wfnCaler=wfn.Calculator(self.mol)
        chars=[]
        molDens=wfnCaler.molDens(pos)
        print(np.min(molDens),np.max(molDens)),print(np.sum(molDens*wei))
        proDens=np.zeros_like(molDens)
        for _,atom in enumerate(self.mol.atoms):
            radius=np.linalg.norm(pos-atom.coord,axis=1)
            proDens+=radDens.get_radDens(atom.atomic,radius)
        print(np.min(proDens),np.max(proDens)),print(np.sum(proDens*wei))
        pass
        for _,atom in enumerate(self.mol.atoms):
            atmDens=wfnCaler.atmDens(atom.idx,pos)
            ratio=np.divide(atmDens,proDens,out=np.zeros_like(atmDens),where=proDens!=0)
            dens=ratio*molDens
            char=sum(dens*wei)
            chars.append(char)
        chars=np.array(chars)
        k=sum(self.mol.atoms.atomics)/sum(chars)
        chars*=k
        chars=[atom.atomic-char for atom,char in zip(self.mol.atoms,chars)]
        return np.array(chars)/20
    
    def dirCharge(self,chrg:Chrgs,atms:list[int],dirs:list[np.ndarray]=None)->np.ndarray:
        """计算不同方向的电荷[n,5](atm,x,y,z,val)"""
        atms_,dirs_=fit_dirs(self.mol,atms,dirs)
        assert len(atms_)==len(dirs_),"长度需要一致"
        dirVal=np.zeros(shape=(len(dirs_),5))
        obts=self.mol.O_obts
        for d in range(len(dirs_)):
            atm=atms_[d]
            dir_=dirs_[d]
            CMp=self.mol.projCM(obts,[atm],[dir_],False,False) # 获取投影后的轨道系数
            PMp=CM2PM(CMp,obts,self.mol.oE)
            if chrg=='mulliken':
                val=self.mulliken(num=True,PM=PMp)[atm-1]
            elif chrg=='lowdin':
                val=self.lowdin(num=True,PM=PMp)[atm-1]
            x,y,z=dir_
            dirVal[d]=[atm,x,y,z,val]

        return dirVal

    def resStr(self)->str:
        """获取结果的打印内容"""
        satoms=lutils.atomIdxs(self.mol.atoms)
        charges=self.calculate()
        return lutils.atomValueStr(self.mol,satoms,charges)
    
    def onShell(self):
        from pywfn.utils import parse_atmList
        printer.info('1. mulliken电荷')
        printer.info('2. lowdin电荷')
        printer.info('3. 方向性电荷')
        while True:
            opt=input('请选择电荷类型:')
            if opt=='':break
            if opt=='1':
                charges=self.mulliken()
                for i,v in enumerate(charges):
                    print(f'{i+1}: {v}')
            if opt=='2':
                charges=self.lowdin()
                for i,v in enumerate(charges):
                    print(f'{i+1}: {v}')
            if opt=='3':
                printer.info('1. Mulliken[*]; 2. lowdin')
                opt=input('请选择方向性电荷类型:')
                if opt not in ['1','2']:return
                if opt=='1':charg='mulliken'
                if opt=='2':charg='lowdin'
                numStr=input('请输入要计算的原子索引：')
                atms=parse_atmList(numStr)
                charges=self.dirCharge(charg,atms)
                for a,x,y,z,v in charges:
                    print(f'{a}: {x},{y},{z}  {v}')
    
def fit_dirs(mol:Mol,atms:list[int],dirs:list[np.ndarray]):
    """
    矫正方向，如果没有指定方向的话，计算每个原子可能的反应方向
    """
    if dirs is None:
        from pywfn.atomprop import direction
        dirCaler=direction.Calculator(mol)
        atms_=[]
        dirs_=[]
        for atm in atms:
            resDirs=dirCaler.reaction(atm)
            dirs_.append(resDirs)
            atms_+=[atm]*len(resDirs)
        dirs_=np.vstack(dirs_)
    else:
        atms_=atms
        dirs_=dirs
    return atms_,dirs_


"""
将分子对象保存为gif文件
"""
from pathlib import Path
import json

from pywfn.base import Mol
from pywfn import config
from pywfn.data import temps

class gjfWriter:
    def __init__(self,mol:Mol):
        self.mol:Mol=mol
        self.temp=temps.gjf
        self.chkPath='chk'
        self.charge=0
        self.multi=1
        
    def get_coordStr(self):
        """生成坐标的字符串形式"""
        coordStrs=[]
        for atom in self.mol.atoms:
            x,y,z=atom.coord
            s=atom.symbol
            coordStr=f'{s:>2}{x:>14.8f}{y:>14.8f}{z:>14.8f}'
            coordStrs.append(coordStr)
        return '\n'.join(coordStrs)
    
    def search(self):
        """搜索需要的各种属性"""
        if self.mol.charge is not None:
            self.charge=self.mol.charge
        if self.mol.spin is not None:
            self.multi=self.mol.spin

    def build(self)->str:
        self.search()
        self.temp=self.temp.replace('<COORD>',self.get_coordStr())
        self.temp=self.temp.replace('<CHARGE>',f'{self.charge}')
        self.temp=self.temp.replace('<MULTI>',f'{self.multi}')
        self.temp=self.temp.replace('<CHK>',self.chkPath)
        return self.temp

    def save(self,path:str):
        self.build()
        Path(path).write_text(self.temp)


"""
定义mol文件的写出器
"""
from pywfn.base import Mol
from pywfn.data import temps
class MolWriter:
    def __init__(self,mol:Mol):
        self.mol = mol
        self.temp=temps.mol
        self.title:str=None

    def atomBlock(self)->str:
        atomLines=[]
        for atom in self.mol.atoms:
            x,y,z=atom.coord/1.889
            s=atom.symbol
            line=f'{x:>10.4f}{y:>10.4f}{z:>10.4f}{s:>2}   0  0  0  0  0  0  0  0  0  0  0  0'
            atomLines.append(line)
        return '\n'.join(atomLines)
    
    def bondBlock(self)->str:
        bondLines=[]
        for bond in self.mol.bonds:
            a1,a2=bond.ats
            line=f'{a1:>3d}{a2:>3d}{bond.btype:>3d}  0  0  0  0'
            bondLines.append(line)
        return '\n'.join(bondLines)
    
    def build(self)->str:
        natom=self.mol.atoms.num
        nbond=self.mol.bonds.num
        if self.title is None:self.title=self.mol.formula
        self.temp=self.temp.replace('[TITLE]',self.title)
        self.temp=self.temp.replace('[NATOM]',f'{natom:>3d}')
        self.temp=self.temp.replace('[NBOND]',f'{nbond:>3d}')
        self.temp=self.temp.replace('[ATOMBLOCK]',self.atomBlock())
        self.temp=self.temp.replace('[BONDBLOCK]',self.bondBlock())
        return self.temp
    
    def save(self,path:str):
        """
        保存mol文件到指定路径
        path:str 文件路径
        """
        with open(path,'w') as f:
            f.write(self.temp)



"""
基础的分子对象，其属性应该是标准的，已知的
一个分子对象应该有哪些属性？基本属性(必须属性),计算属性(需要计算才能的到的属性)
结构相关
- 原子
- 键
- 法向量
- 重原子

分子轨道相关
轨道数量
轨道类型
是否分为α和β
占据轨道序数
非占据轨道序数
重叠矩阵

读取器
"""

from typing import Any
from pywfn import maths
from pywfn.base.atom import Atom,Atoms
from pywfn.base.bond import Bond,Bonds
from pywfn import reader
from pywfn import data
from pywfn.data.elements import elements
from pywfn import config
from pywfn import utils
from pywfn import reader
from pywfn.utils import printer

import numpy as np
from functools import cached_property, lru_cache
from tqdm import tqdm
import multiprocessing as mp
from multiprocessing.pool import AsyncResult
import threading
from typing import Callable
import collections
import re

class Props(dict):
    def __init__(self) -> None:
        super().__init__({})

    def get(self,key:str,fun:Callable):
        if key not in self.keys():
            self[key]=fun()
        return self[key]

    def set(self,key:str,value):
        self[key]=value

class Mol:
    """基础的分子对象"""
    def __init__(self,reader:"reader.Reader"=None) -> None:
        """
        分子实例化
        `reader:Reader`，分子读取器
        """
        self._atoms:Atoms=Atoms(self)
        self._bonds:Bonds=Bonds(self)
        self.reader:"reader.Reader"=reader
        self.props=Props()
        self.bohr:bool=False # 是否使用波尔坐标
    
    @cached_property
    def basis(self)->data.Basis: # 设为属性，可以保证用不到的时候不被实例化
        return self.reader.get_basis()
    
    @cached_property
    def gto(self)->"maths.Gto":
        return maths.Gto(self)

    @cached_property
    def charge(self)->int:
        return self.reader.get_charge()
    
    @cached_property
    def spin(self)->int:
        return self.reader.get_spin()

    @property
    def open(self)->bool:
        """是否为开壳层"""
        w,h=self.CM.shape
        return w!=h
    
    @property
    def energy(self)->float:
        """获取分子能量"""
        return self.props.get('energy',self.reader.get_energy)

    @property
    def obtOccs(self)->list[bool]:
        """获取每个分子轨道是否占据"""
        occs=self.props.get('obtOccs',self.reader.get_obtOccs)
        return occs

    @cached_property
    def obtEngs(self)->list[float]:
        return self.reader.get_obtEngs()

    @cached_property
    def obtStrs(self)->list[str]:
        """返回轨道符号"""
        strs=[]
        occs=self.obtOccs
        nobt=len(occs)
        for i,s in enumerate(occs):
            if self.open:
                if i<len(nobt)//2:
                    s=f'α {s}'
                else:
                    s=f'β {s}'
            idx=i
            if self.open and i>=nobt//2:
                idx=i-nobt//2
            strs.append(f'{idx+1} {s}')
        return strs
    
    @cached_property
    def obtAtms(self)->list[int]:
        return self.reader.get_obtAtms()
    
    @cached_property
    def obtShls(self)->list[int]:
        return self.reader.get_obtShls()

    @cached_property
    def obtSyms(self)->list[str]:
        return self.reader.get_obtSyms()

    @property
    def atoms(self)->Atoms:
        """获取所有原子"""
        if self._atoms:return self._atoms
        symbols=self.reader.get_symbols()
        coords=self.reader.get_coords()
        for s,c in zip(symbols,coords):
            self._atoms.add(symbol=s,coord=c)
        return self._atoms
    
    @property
    def bonds(self)->Bonds:
        if self._bonds:return self._bonds
        for atom1 in self.atoms:
            for atom2 in self.atoms:
                if atom1.idx>=atom2.idx:continue
                r=np.linalg.norm(atom2.coord-atom1.coord)
                # r_=elements[atom1.symbol].radius+elements[atom2.symbol].radius
                r_=1.7*1.889
                if r>r_:continue
                self._bonds.add(atom1.idx,atom2.idx)
        return self._bonds
    
    def atom(self,idx:int)->Atom:
        """根据原子编号获取一个原子，从1开始"""
        assert isinstance(idx,int),f"索引应该为整数"
        assert idx>0,"索引从1开始"
        assert idx<=len(self.atoms),"原子索引超过原子数量"
        return self.atoms[idx-1]
    
    @lru_cache
    def bond(self,idx1:int,idx2:int)->Bond:
        """第一次调用是生成键,第二次调用时直接返回,秒啊"""
        return self.bonds.get(idx1,idx2)

    @cached_property
    def coords(self)->np.ndarray:
        """返回原子坐标矩阵[n,3]"""
        return np.array([atom.coord for atom in self.atoms])*1.889

    @property
    def O_obts(self)->list[int]:
        return [i for i,occ in enumerate(self.obtOccs) if occ]
    
    @cached_property
    def V_obts(self)->list[int]:
        return [i for i,occ in enumerate(self.obtOccs) if not occ]
    
    @cached_property
    def heavyAtoms(self)->list[int]:
        return [atom.idx for atom in self.atoms if atom.symbol!='H']

    @property
    def CM(self)->np.ndarray:
        """分子轨道系数矩阵"""
        return self.props.get('CM',self.reader.get_CM)
    
    @property
    def SM(self):
        """重叠矩阵"""
        return self.reader.get_SM()
    
    @property
    def PM(self):
        """密度矩阵"""
        return maths.CM2PM(self.CM.copy(),self.O_obts,self.oE)
    
    @property
    def oE(self):
        """轨道电子数"""
        return 1 if self.open else 2
    
    @property
    def formula(self)->str:
        """获取分子式"""
        symbols=self.atoms.symbols
        # 统计列表中每个元素出现的次数
        counts = collections.Counter(symbols)
        names=[f'{k}{v}' for k,v in counts.items()]
        return ''.join(names)
    

    def params(self,atms:list[int])->float:
        """
        获取键长、键角、二面角
        """
        if len(atms)==2:
            a1,a2=atms
            return self.bond(a1,a2).length
        elif len(atms)==3:
            a1,a2,a3=atms
            v21=self.atom(a1).coord-self.atom(a2).coord
            v23=self.atom(a3).coord-self.atom(a2).coord
            angle=maths.vector_angle(v21,v23)
            return angle
        elif len(atms)==4:
            a1,a2,a3,a4=atms
            v21=self.atom(a1).coord-self.atom(a2).coord
            v23=self.atom(a3).coord-self.atom(a2).coord
            v32=self.atom(a2).coord-self.atom(a3).coord
            v34=self.atom(a3).coord-self.atom(a4).coord
            n2=np.cross(v21,v23)
            n3=np.cross(v34,v32)
            n2/=np.linalg.norm(n2)
            n3/=np.linalg.norm(n3)
            angle=maths.vector_angle(n2,n3)
            nm=np.dot(v21,n3) # 为了实现二面角的正负而引入
            nm/=np.linalg.norm(nm)
            return angle/nm*np.pi
        else:
            raise ValueError("参数数量错误")

    
    def projCM(self,obts:list[int],atms:list[int],dirs:list[np.ndarray]
               ,akeep:bool,lkeep:bool,keeps:str=None)->np.ndarray:
        """
        获取投影后的系数矩阵
        atms:需要投影的原子
        obts:需要投影的轨道
        dirs:投影到的方向 atms和dirs的长度必须相同
        akeep:其它原子系数是否保留 keep other atoms
        lkeep:其它价层系数是否保留 keep other layer
        keeps:额外保留的轨道，可以使用正则表达式匹配
        """
        assert isinstance(dirs,list),"方向想两需要为列表"
        assert len(atms)==len(dirs),"原子和方向数量不同"
        if akeep:
            CMp=np.copy(self.CM)
        else:
            CMp=np.zeros_like(self.CM,dtype=np.float32) #新的系数矩阵
            
        for a,(atom,vect) in enumerate(zip(atms,dirs)):
            atom=self.atom(atom)
            nebNum=len(atom.neighbors)
            u,l=atom.obtBorder
            syms=self.obtSyms[u:l]
            
            pIdx=[i for i,s in enumerate(syms) if 'P' in s]
            if len(pIdx)==0:continue # 没有p轨道则跳过
            for o,obt in enumerate(obts):
                if lkeep:
                    Co=self.CM.copy()[u:l,obt]
                else:
                    Co=np.zeros(len(syms)) #根据是否P轨道之外的保留还是0由不同的选择
                if keeps is not None: # 其它保留的层
                    idxs=[i for i,s in enumerate(syms) if re.match(s)]
                    Cos=atom.obtCoeffs.copy()[idxs,obt]
                    Co[idxs]=Cos
                Cop=atom.get_pProj(vect,obt)
                Co[pIdx]=np.concatenate(Cop)
                CMp[u:l,obt]=Co.copy()
        return CMp

    def __repr__(self):
        return f'atom number: {len(self.atoms)}'

    def get_wfnv(self,obt:int,pos:np.ndarray,atoms:list[int])->np.ndarray:
        """
        导出指定点的分子轨道数据
        obt: 要渲染的分子轨道
        pos: 要渲染的格点坐标[n,3]
        atoms: 要渲染的原子，索引从1开始
        """
        print(f'渲染轨道{obt=},{pos.shape},{atoms}')
        assert pos.shape[1]==3,'坐标的形状需为[n,3]'
        values=np.zeros(len(pos))
        # pool=mp.Pool(mp.cpu_count())
        # proces:list[AsyncResult]=[]
        # multi=True
        # if multi:
        #     for idx in atoms:
        #         atom=self.atom(idx)
        #         posi=pos-atom.coord
        #         proce=pool.apply_async(atom.get_wfnv,(posi,obt,))
        #         proces.append(proce)
        #     for proce in tqdm(proces):
        #         values+=proce.get()
        #     return values
        # else:
        for idx in atoms:
            atom=self.atom(idx)
            posi=pos-atom.coord
            values+=atom.get_wfnv(posi,obt)
        return values
    
    def get_dens(self,atoms:list[int],obts:list[int],coords:np.ndarray)->np.ndarray:
        """
        计算分子的电子密度
        atoms:要计算的原子，从1开始
        coords:空间笛卡尔坐标
        """
        molDens=np.zeros(shape=(len(coords))) # 每一个坐标的电子密度值
        for a in atoms:
            atom=self.atom(a)
            coord=coords-atom.coord #以原子为中心的坐标
            atmDens=atom.get_dens(obts,coord)
            molDens+=atmDens
        return molDens


"""
传入一堆位置点，计算在这些位置的各种函数数值
传入的数组大小为[n,3],一行分别是x,y,z
基组的种类是有限的吗？有没有通用的表达形式？为每一个基组都定义不同的计算类型
将多次的向量计算转为多次的矩阵运算
坐标数量 N
收缩数量 S
原子轨道数量 L
原子数量 N
一个gto可以分为两部分

一个原子轨道由许多电子轨道组成，每个电子轨道最多容纳两个电子
每个电子轨道由多个gto组成，用多个gto去拟合(线性组合)一个电子轨道

每一个电子轨道有一个对应的分子轨道系数

电子轨道又可以分为不同层

每一层有不同的角动量 l+m+n
每一个角动量又有三个分量 l,m,n
每个电子轨道对应一个角动量分量
组成电子轨道的gto的系数相同但指数不同c1*gto(a)+c2*gto(a)+c3*gto(a)

代码的可读性和速度如何取舍？ 我还是选择了后者

可以将gto的计算放到多个进程中进行

分子轨道波函数
原子轨道波函数
原子波函数->原子电子密度
"""

import numpy as np
from numpy import sqrt, ndarray

from pywfn import base
from pywfn.utils import printer
from pywfn.data import Basis

π = np.pi
e = np.e


def fac2(num):  # 该例中一定是奇数
    """计算双阶乘"""
    # num+=1
    if num <= 1:
        fac = 1
    else:
        fac = np.prod(np.arange(1, num + 1, 2))
    return fac


class Gto:
    def __init__(self, mol: "base.Mol"):
        self.mol = mol
        self.basis = mol.basis
        self.angs = [0, 1, 2, 3, 4]  # 默认全部绘制

    def ato(self, pos: ndarray, atm: int, obts: list[int]):
        """
        计算原子的高斯型波函数数值
        pos:坐标，以原子为中心
        atm:原子
        obt:轨道
        """
        assert len(pos.shape) == 2, f"pos的维度应该为2"
        assert pos.shape[1] == 3, f"pos的形状应为[n,3]，当前为{pos.shape}"
        R2 = np.sum(pos**2, axis=1)  # x^2+y^2+z^2

        atom = self.mol.atom(atm)
        u, l = atom.obtBorder

        atms = self.mol.obtAtms[u:l]
        shls = self.mol.obtShls[u:l]
        syms = self.mol.obtSyms[u:l]
        lmns = [Basis.sym2lmn(sym) for sym in syms]
        coef = self.mol.CM[u:l, obts]  # 二维矩阵
        expl = []
        coel = []
        ncsl = []  # 记录每个收缩轨道的大小
        for i in range(len(atms)):  # 该原子的行索引
            lmn = lmns[i]
            atm = atms[i]
            shl = shls[i]
            ang = sum(lmn)
            basis = self.mol.basis.get(atom.atomic, shl, ang)
            exps = [b.exp for b in basis]
            coes = [b.coe for b in basis]
            expl += exps
            coel += coes
            ncsl.append(len(exps))
        wfns = Gto.agf(expl, coel, ncsl, lmns, R2, pos, coef, obts)
        return wfns

    @staticmethod
    def agf(
        expl: list[float],
        coel: list[float],
        ncsl: list[int],
        lmns: list[int],
        R2: np.ndarray,
        pos: np.ndarray,
        coef: np.ndarray,
        obts: list[int],
    ) -> np.ndarray:
        nexp = len(expl)
        j = 0
        
        wfns = np.zeros(len(pos), dtype=np.float32)
        for i in range(len(lmns)):
            nc = ncsl[i]  # 收缩大小
            lmn = lmns[i]
            coes = coel[j : j + nc]
            exps = expl[j : j + nc]
            j += nc
            wfn = Gto.cgf(exps, coes, lmn, R2, pos) # 每一个原子轨道都是提前定义好的不变的
            for obt in obts:
                wfns += coef[i, obt] * wfn
        return wfns

    @staticmethod
    def cgf(
        exps: list[float],
        coes: list[float],
        lmn: list[int],
        R2: np.ndarray,
        pos: np.ndarray,
    ) -> np.ndarray:
        """
        收缩高斯函数，线性组合之前的波函数
        pos:坐标[n,3]，以原子为中心
        c1*f1+c2*f2...
        """
        # R2=np.sum(pos**2,axis=1)
        wfns = np.zeros(len(pos))
        for exp, coe in zip(exps, coes):
            wfns += coe * Gto.gtf(exp, R2, pos, lmn)
        return wfns

    @staticmethod
    def gtf(exp: float, R2: ndarray, pos: ndarray, lmn: list[int]) -> ndarray:
        """
        计算指定点gto函数的值
        exp:高斯指数
        pos:坐标[n,3]，以原子为中心
        R:坐标到原点距离平方
        lmn:角动量决定的x,y,z指数
        """
        # i=0,(2i-1)=-1,(2i-1)!!=1
        # i=1,(2i-1)= 1,(2i-1)!!=1
        # i=2,(2i-1)= 3,(2i-1)!!=3
        facs=[1,1,3] # n与阶乘的对应关系？
        l, m, n = lmn
        x, y, z = pos.T
        ang = sum(lmn)  # 角动量
        fac = facs[l] * facs[m] * facs[n]  # 双阶乘
        N = (2 * exp / π) ** (3 / 4) * np.sqrt((4 * exp) ** ang / fac)
        # N = (2 * exp / π) ** (3 / 4) * np.sqrt((2 ** (2 * ang) * exp**ang) / fac)
        wfn = x**l * y**m * z**n * np.exp(-exp * R2) * N
        return wfn



"""
一个原子的基组应该有一下层级

不同价层
    指数(一维数组)
    系数(二维数组)
不同角动量对应不同的mnl
所有可能的m+n+l=角动量决定了基函数的数量

"""

from dataclasses import dataclass
from pywfn import utils

printer = utils.Printer()

from functools import lru_cache


@dataclass
class BasisData:
    atmic: int  # 元素
    shl: int  # 壳层
    ang: int  # 角动量
    exp: float  # 指数
    coe: float  # 系数

    def __iter__(self):
        data = self.atmic, self.shl, self.ang, self.exp, self.coe
        return iter(data)


class Basis:
    """
    所有提前准备的基组数据
    """

    def __init__(self, name: str) -> None:
        """根据基组名实例化基组信息"""
        self.name = name
        self.data: list[BasisData] = None

    def atomics(self):
        atmics = [basis.atmic for basis in self.data]
        return list(set(atmics))

    def setData(self, data: list[BasisData]):
        """设置数据
        元素,层数,角动量,指数,系数
        idx,shl,ang,exp,coe
        """
        self.data = data

    @lru_cache
    def get(self, atmic: int, shell: int = None, ang: int = None) -> list[BasisData]:
        """根据原子序号获得基组"""
        if shell is None:
            basis = [b for b in self.data if b.atmic == atmic]
        else:
            basis = [
                b
                for b in self.data
                if (b.atmic == atmic and b.shl == shell and b.ang == ang)
            ]
        assert len(basis) > 0, "没有基组信息"
        return basis

    @staticmethod
    def ang2lmn(ang: int) -> list[list[int]]:
        """
        根据角动量获取l,m,n
        基函数中包含 x^l.y^m.z^n
        注意，角动量要与计算时的系数对应
        """
        lmns = {
            0: [[0, 0, 0]],
            1: [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            2: [[2, 0, 0], [0, 2, 0], [0, 0, 2], [1, 1, 0], [1, 0, 1], [0, 1, 1]],
        }
        return lmns[ang]

    @staticmethod
    def sym2lmn(sym: str)->list[int]:
        lmnMap = {
            "S": [0, 0, 0],
            "PX": [1, 0, 0],
            "PY": [0, 1, 0],
            "PZ": [0, 0, 1],
            "XX": [2, 0, 0],
            "YY": [0, 2, 0],
            "ZZ": [0, 0, 2],
            "XY": [1, 1, 0],
            "XZ": [1, 0, 1],
            "YZ": [0, 1, 1],
        }
        return lmnMap[sym]
    
    @staticmethod
    def sym2Ang(sym: str)->list[int]:
        lmn=Basis.sym2lmn(sym)
        return sum(lmn)

    def num(self, atomic: int) -> int:
        """获取基组原子对应的基函数数量"""
        data = self.get(atomic)
        return sum([len(self.lmn(ang)) for ang, _, _ in data])

    def lName(self, l, m, n):
        key = f"{l}{m}{n}"
        names = {
            "000": "S",
            "100": "PX",
            "010": "PY",
            "001": "PZ",
            "200": "XX",
            "020": "YY",
            "002": "ZZ",
            "110": "XY",
            "101": "XZ",
            "011": "YZ",
        }
        return names[key]

    def numAng(self, strs):
        """将角动量符号转为数值"""
        return [[int(i) for i in s] for s in strs]



"""
输入文件有很多结构，将每个结构分割来来
重点在于分割，首先根据指定的分割符将文本分割为很多块，然后在每一块提取所需内容
"""
import re,os
from typing import Any
from pathlib import Path

from pywfn.tools import writer
from pywfn.data import Elements
from pywfn.utils import printer

class Tool:
    def __init__(self,path:str) -> None:
        """输入log文件的路径"""
        self.path=Path(path)
        self.dirName=self.path.parent
        self.baseName=self.path.name # 包含后缀的文件名
        self.fileName=self.path.stem # 不包含后缀的文件名
        folder=self.dirName / self.fileName
        if not Path(folder).exists(): #判断文件夹是否存在
            os.mkdir(folder) # 创建文件夹
        self.content=self.path.read_text(encoding='utf-8')
        self.lines=self.content.splitlines(keepends=False)

        if 'Standard orientation' in self.content: # 坐标类型，是否为标准朝向
            self.coordType='Standard orientation'
        else:
            self.coordType='Input orientation'
        self.elements=Elements()

    def split_raw(self)->list[str]:
        """将原始的一大段文本分割"""
        mark=' Optimization completed.' # 以此为每个结构的分隔，因为是柔性扫描，因此每一步结束都有这么一个标志
        # blocks=self.content.split(mark)
        blocks=re.split(mark,self.content)
        return blocks
    
    def find_titles(self,blocks:list[str]):
        """找到坐标对应的标题行"""
        titleNums=[]
        for idx,block in enumerate(blocks): # 循环每一段
            titleNum=[idx,0]
            lines=block.splitlines(keepends=False)
            for i,line in enumerate(lines): # 循环每一行
                
                if self.coordType in line: # 找到行坐标
                    titleNum=[idx,i]
                    # titleNums.append((idx,i)) # 只要第一个,记录段落数和行数，为啥只要第一个？不应该要最后一个？
                    # break
            titleNums.append(titleNum)
        return titleNums
    
    def match_coords(self,blocks:list[str],titleNums:list[tuple[int]])->list[list[float]]:
        """返回包含全部结构坐标的三维数组"""
        s=r" +\d+ +(\d+) +\d+ +(-?\d+.\d+) +(-?\d+.\d+) +(-?\d+.\d+)"
        coords=[]
        for idx,titleNum in titleNums:
            coord=[]
            lines=blocks[idx].splitlines(keepends=False)
            for i in range(titleNum+5,len(lines)): # 从标题后的第五行开始
                line=lines[i]
                if re.search(s,line) is not None:
                    idx,x,y,z=re.search(s,line).groups()
                    symbol=self.elements[int(idx)].symbol
                    coord.append([symbol,float(x),float(y),float(z)])
                else:
                    break
            coords.append(coord)
        return coords
            
    def split(self):
        """分割文件"""
        blocks=self.split_raw()
        # 首先获取所有构象的坐标

        titleNums=self.find_titles(blocks)
        coords=self.match_coords(blocks,titleNums)
        for i,coord in printer.track(enumerate(coords)):
            path=os.path.join(self.dirName,self.fileName,f'f{i+1:0>2}.gjf')
            writer.gjf(path,f'{i+1}',coord,0,1)

    def __call__(self) -> Any:
        self.blocks=self.split_raw()



"""
本脚本用以生成势能面
"""
from pathlib import Path

from pywfn.base import Mol
from pywfn.reader import get_reader,LogReader
from pywfn.data import temps
from collections import namedtuple
from dataclasses import dataclass
from typing import Union

@dataclass
class Mole:
    energy:float # 分子能量
    name:str='' # 分子名

@dataclass
class Moles:
    mols:list[Mole]
    text:str='mol'
    n0:Union[None,"Node"]=None # 开始的节点
    n1:Union[None,"Node"]=None # 结束的节点
    posY:float=0
    @property
    def energy(self):
        return sum([mol.energy for mol in self.mols])
    

@dataclass
class Block:
    blocks:"Blocks"
    idx:int
    Rms:Moles # 反应物
    Tms:Moles # 过渡态
    Pms:Moles # 产物
    bi0:Union[int,"Block"]=None
    bi1:Union[int,"Block"]=None

    @property
    def b0(self)->"Block":
        if self.bi0 is None:
            return None
        else:
            return self.blocks[self.bi0]

    @property
    def b1(self)->"Block":
        if self.bi1 is None:
            return None
        else:
            return self.blocks[self.bi1]

    @property
    def engs(self)->list[float]:
        e0=self.Rms.energy
        es=[self.Rms.energy,self.Tms.energy,self.Pms.energy]
        es=[e-e0 for e in es] # 反应物为0的能量
        if self.b0:
            es=[e+self.b0.Pms.energy for e in es] # 接着上一块的能量
        return es
    
    @property
    def pos(self):
        if self.b0:
            return self.b0.pos+Sizes.block
        else:
            return 0
    
    def __repr__(self):
        return f'{self.bi0}<==({self.idx})==>{self.bi1}'

@classmethod
class Blocks:
    blocks:list[Block]=[]

    def __getitem__(self,idx)->Block:
        for block in self.blocks:
            if idx==block.idx:
                return block
    
    def append(self,block:Block):
        self.blocks.append(block)

@dataclass
class Node: # chemdraw中的一个节点
    id:int
    x:float
    y:float

@dataclass
class Bond:
    id:int
    B:int
    E:int
    Display:str

@dataclass
class Text:
    id:int
    t:str
    x:float
    y:float

class Sizes:
    block:float=90 # 一个块的大小
    moles:float=30 # 一个分子组的大小
    startx:float=50
    starty:float=100


class Tool:
    def __init__(self) -> None:
        self.blocks:list[Block]=[]
        # self.route:list[int]
        self.temp=temps.pes
        self.nodes:list[Node]=[]
        self.bonds:list[Bond]=[]
        self.texts:list[Text]=[]
        self.NBS=''
        self.TXS=''
        self.idx=4
    
    def getIdx(self):
        self.idx+=1
        return self.idx
    
    def getPos(self,x:float,y:float):
        return x+Sizes.startx+Sizes.block/2+10,400-(y+Sizes.starty)

    def create(self):
        for b,block in enumerate(self.blocks):
            self.add_block(block)
        self.build_bonds()
        self.write()

    def add_block(self,block:Block):
        """
        添加一个Block块
        bid:第多少个块
        """
        Rms=block.Rms
        Tms=block.Tms
        Pms=block.Pms
        eng0=0 if block.b0 is None else block.b0.Pms.posY
        eng1=eng0+(Tms.energy-Rms.energy)
        eng2=eng0+(Pms.energy-Rms.energy)
        Rms.posY,Tms.posY,Pms.posY=eng0,eng1,eng2
        
        
        x1=block.pos
        x2=x1+Sizes.block/2
        
        if block.b0 is None:
            x0=block.pos-Sizes.block/2
            Rms.n0,Rms.n1=self.add_moles(eng0,x0,Rms.text)
            
        Tms.n0,Tms.n1=self.add_moles(eng1,x1,Tms.text)
        Pms.n0,Pms.n1=self.add_moles(eng2,x2,Pms.text)
        print(block.idx,x1)
        

    def add_moles(self,eng:float,x:float,text:str): # 一个moles是一个横线
        x0=x
        x1=x+Sizes.moles
        n0=self.add_node(x0,eng)
        n1=self.add_node(x1,eng)
        self.add_text(text,(x0+x1)/2-len(text)*4,eng+2)
        return n0,n1
            
    def add_node(self,x,y)->Node:
        """添加一个节点"""
        x,y=self.getPos(x,y)
        node=Node(id=self.getIdx(),x=x,y=y)
        self.nodes.append(node)
        return node

    def build_bonds(self):
        """生成键"""
        for block in self.blocks:
            if block.b0 is None:
                self.add_bond(block.Rms.n0,block.Rms.n1,'Bold')
                self.add_bond(block.Rms.n1,block.Tms.n0,'Dash')
            if block.b0:
                self.add_bond(block.b0.Pms.n1,block.Tms.n0,'Dash')
            self.add_bond(block.Tms.n0,block.Tms.n1,'Bold')
            self.add_bond(block.Tms.n1,block.Pms.n0,'Dash')
            self.add_bond(block.Pms.n0,block.Pms.n1,'Bold')
            
    def add_bond(self,nb:Node,ne:Node,display:str):
        """添加键"""
        bond=Bond(id=self.getIdx(),B=nb.id,E=ne.id,Display=display)
        self.bonds.append(bond)
        return self.idx
    
    def add_text(self,text:str,x:float,y:float):
        """添加文本"""
        x,y=self.getPos(x,y)
        self.texts.append(Text(id=self.getIdx(),t=text,x=x,y=y))
        
    
    def write(self):
        for node in self.nodes:
            self.NBS+=f'<n id="{node.id}" p="{node.x} {node.y}" Z="1" AS="N"/>\n'
        for bond in self.bonds:
            self.NBS+=f'<b id="{bond.id}" B="{bond.B}" E="{bond.E}" Z="2" Display="{bond.Display}" BS="N"/>\n'
        for text in self.texts:
            self.TXS+=f'<t id="{text.id}" p="{text.x} {text.y}" Z="3" LineHeight="auto"><s font="5" size="10" color="0">{text.t}</s></t>\n'
        self.temp=self.temp.replace('[NBS]',self.NBS)
        self.temp=self.temp.replace('[TXS]',self.TXS)
        self.temp=self.temp.replace('[BOUND]','0 0 10 20')
        (Path.cwd()/'pes.cdxml').write_text(self.temp)
        # print(self.temp)

    


"""
轨道投影法+Mayer计算π键级
根据投影计算新的分子轨道系数矩阵,然后重构密度矩阵
"""

from pywfn.base import Mol
from typing import *
import numpy as np

from pywfn import maths
from pywfn.bondprop import Caler
from pywfn.utils import printer

class Calculator(Caler):
    def __init__(self,mol:Mol) -> None:
        self.mol=mol
        self.values={}
        self.vect:Union[None,np.ndarray]=None
        self.bond:list[int,int]=None
        self.zero:bool=True
        self.keep:bool=True
        self.ins:bool=True

    def calculate(self)->np.ndarray:
        """
        计算pi键级，当指定方向时计算方向键级
        """
        assert self.bond is not None,'没有指定键级'
        idx1,idx2=self.bond
        printer.info(f'计算{idx1} → {idx2}的键级')
        
        centerAtom=self.mol.atom(idx1)
        aroundAtom=self.mol.atom(idx2)
        if centerAtom.symbol=='H' or aroundAtom.symbol=='H':
            return 0.,0.
        obts=self.mol.O_obts #占据轨道的索引

        vect=self.vect
        if vect is None:
            normal=centerAtom.get_Normal(aroundAtom.idx)
        else:
            normal=vect/np.linalg.norm(vect)
        
        
        if len(centerAtom.neighbors)==3:
            printer.vector('法向方向: ',normal)
            return self.calerWay(obts,normal),0.
        elif len(centerAtom.neighbors) in [2,1]:
            bondVector=aroundAtom.coord-centerAtom.coord
            if normal is None:
                idxn=maths.search_sp2(idx1,self.mol)
                if idxn is None:
                    normal=centerAtom.get_vertObt(idx1,idx2)
                    printer.vector('轨道方向: ',normal)
                else:
                    normal=self.mol.atom(idxn).get_Normal(idx2)
                    printer.vector('相邻法向: ',normal)
            else:
                printer.vector('法向方向: ',normal)
            
            
            cross=np.cross(bondVector,normal)
            cross/=np.linalg.norm(cross)
            printer.vector('交叉方向: ',cross)
            vn,vc=self.calerWay(obts,normal),self.calerWay(obts,cross)
            return np.array([vn,vc])
        else:
            return np.array([0.,0.])

    def calerWay(self,obts:list[int],norm:np.ndarray):
        """计算某个方向的键级"""
        norm.flags.writeable=False
        length=np.linalg.norm(norm)
        assert abs(length-1)<1e-4,"向量长度应为1"
        idx1,idx2=self.bond
        """计算两个原子之间某个方向的键级"""
        atom1=self.mol.atom(idx1)
        atom2=self.mol.atom(idx2)
        a_1,a_2=atom1.obtBorder
        b_1,b_2=atom2.obtBorder
        atoms=[idx1,idx2]
        vects=[norm]*len(atoms)
        CM_=self.mol.projCM(atoms,obts,vects,zero=self.zero,keep=self.keep)

        oe=self.mol.oE
        SM=self.mol.SM
        
        PM_=maths.CM2PM(CM_,obts,oe)
        PS=PM_@SM
        order=np.sum(PS[a_1:a_2,b_1:b_2]*PS[b_1:b_2,a_1:a_2].T)
        return np.sqrt(np.abs(order))
    

    def resStr(self) -> str:
        order1,order2=self.calculate()
        return f'{order1:.4f},{order2:.4f}'
    
    def print(self):
        printer.res(self.resStr())


"""
计算原子的各种方向
- 原子法向量
- p轨道方向
- 原子周围一圈方向
"""
from pywfn.base import Mol
from pywfn import maths
from pywfn.maths import vector_angle,points_rotate

import numpy as np

class Calculator:
    def __init__(self,mol:Mol) -> None:
        self.mol=mol

    def pObt(self)->np.ndarray:
        """计算p轨道的方向向量"""
        pass

    def maxWfn(self)->np.ndarray:
        """计算波函数最大值方向"""

    def sphAro(self)->np.ndarray:
        """计算周围一圈的球形范围"""
        pass

    def reaction(self,atm:int)->np.ndarray:
        """ reaction around
        计算原子可能的反应方向 以原子为中心的向量

        - 直线sp原子，垂直于键轴
        - 弯曲sp原子，键轴夹角垂面，且与键轴夹角小于90
        - 平面sp2，垂直于平面的两个方向
        - 弯曲sp2，垂直于平面，且与键轴之间夹角大于90°
        """
        atom=self.mol.atom(atm)
        nebs=atom.neighbors
        if len(nebs)==2:
            ia,ib=nebs
            va=self.mol.atom(ia).coord-atom.coord
            vb=self.mol.atom(ib).coord-atom.coord
            va/=np.linalg.norm(va)
            vb/=np.linalg.norm(vb)

            angle=vector_angle(va,vb) # 两向量之间的夹角
            linear=abs(1-angle)<1e-1 # 是否为线性
            cent=atom.coord
            axis=vb-va
            if linear:
                anyv=np.random.rand(3) # 定义一个任意的向量
                anyv/=np.linalg.norm(anyv)
                cros=np.cross(axis,anyv) # 计算垂直于键轴和任意向量的向量
                cros/=np.linalg.norm(cros)
                angs=np.linspace(0,np.pi*2,35,endpoint=False)
            else:
                cros=np.cross(va,vb) # 垂直于va和vb的向量
                cros/=np.linalg.norm(cros)
                angs=np.linspace(0,np.pi,18,endpoint=True)
            points=(cent+cros).reshape(-1,3)
            dirs=[points_rotate(points,cent,axis,ang)-cent for ang in angs]
            dirs=np.concatenate(dirs,axis=0)
            return dirs
        if len(nebs)==3:
            pa,pb,pc=[self.mol.atom(n).coord for n in nebs]
            params=maths.get_plane_by_3points(pa,pb,pc)
            distan=maths.get_distance_to_plane(params,atom.coord)
            planer=distan<1e-1 # 是否为平面
            a,b,c,_=params
            normal=np.array([a,b,c])
            normal=normal/np.linalg.norm(normal)
            if planer:
                return np.array([normal,-normal])
            else:
                angle=vector_angle(normal,pa-atom.coord)
                if angle>0.5:normal*=-1
                return normal.reshape(1,3)
        
        raise ValueError("找不到可能的反应方向")
    
    def normal(self,atm:int)->np.ndarray|None:
        """计算原子的法向量"""
        atom=self.mol.atom(atm)
        nebs=atom.neighbors
        if len(nebs)==2:
            ia,ib=nebs
            va=self.mol.atom(ia).coord-atom.coord
            vb=self.mol.atom(ib).coord-atom.coord
            angle=vector_angle(va,vb)
            linear=abs(1-angle)<1e-1
            if linear:
                return None
            else:
                va/=np.linalg.norm(va)
                vb/=np.linalg.norm(vb)
                normal=np.cross(va,vb)
                return normal/np.linalg.norm(normal)
        if len(nebs)==3:
            pa,pb,pc=[self.mol.atom(n).coord.copy() for n in nebs]
            pa/=np.linalg.norm(pa)
            pb/=np.linalg.norm(pb)
            pc/=np.linalg.norm(pc)
            vab=pb-pa
            vac=pc-pa
            normal=np.cross(vab,vac)
            return normal/np.linalg.norm(normal)
        return None



"""
轨道挑选法+mayer键级公式
mayer键级需要重叠矩阵,可一次性计算所有键级(此时原子的法向量由三个原子决定,而不是垂直于键轴)
"""
import numpy as np

from pywfn.base import Mol
from pywfn.bondprop import lutils,Caler
from pywfn import maths

class Calculator(Caler):
    def __init__(self,mol:Mol) -> None:
        self.mol=mol
        self.bond:list[int]=None

    def calculate(self)->float:
        """指定两个原子,计算π键键级"""
        idx1,idx2=self.bond
        atom1=self.mol.atom(idx1)
        atom2=self.mol.atom(idx2)
        normal=atom1.get_Normal()
        if normal is None:
            idxn=maths.search_sp2(idx1,self.mol)
            if idxn is None:
                normal=atom1.get_vertObt(atom2.idx,atom1.idx)
            else:
                normal=self.mol.atom(idxn).get_Normal(idx2)

        obts=self.mol.O_obts
        CM_=np.zeros_like(self.mol.CM) # 拷贝一份，然后将不是π轨道的那些变成0

        piObts=[]
        for atom in [atom1,atom2]: # 修改每个原子对应的系数矩阵
            if atom.symbol=='H':continue
            a_1,a_2=atom.obtBorder
            for orbital in obts:
                judgeRes=lutils.judgeOrbital(atom1,atom2,orbital,normal)
                if judgeRes==0:continue # 如果是π轨道
                CM_[a_1:a_2,orbital]=self.mol.CM[a_1:a_2,orbital]
                piObts.append(orbital)
        oe=self.mol.oE
        PM_=lutils.CM2PM(CM_,obts,oe)
        SM=self.mol.SM
        PS=PM_@SM

        a1,a2=atom1.obtBorder
        b1,b2=atom2.obtBorder
        order=np.sum(PS[a1:a2,b1:b2]*PS[b1:b2,a1:a2])
        piObts=set(piObts)
        piObts=[self.mol.obtStrs[o] for o in piObts]
        lutils.formPrint(contents=[piObts],eachLength=10)
        return order